using System;
using System.Text;

namespace Smdn.Protocols.Imap4 {
  public static class ImapEncoding {
    // for string which contains 7-bit characters
    public static readonly Encoding Encoding7Bit = Encoding.ASCII;

    // for string which contains 8-bit characters
    public static readonly Encoding Encoding8Bit = Encoding.GetEncoding("ISO-8859-1");

    // US-ASCII
    private static readonly Encoding USAscii = Encoding.ASCII;

    public static string ToBase64String(string str)
    {
      return Convert.ToBase64String(USAscii.GetBytes(str));
    }

    public static string ToBase64String(string str, Encoding encoding)
    {
      return Convert.ToBase64String(encoding.GetBytes(str));
    }

    public static string FromBase64String(string str)
    {
      return USAscii.GetString(Convert.FromBase64String(str));
    }

    public static string FromBase64String(string str, Encoding encoding)
    {
      return encoding.GetString(Convert.FromBase64String(str));
    }

    internal static string FromModifiedUTF7StringNoException(string str)
    {
      try {
        return FromModifiedUTF7String(str);
      }
      catch (FormatException) {
        return str;
      }
    }

    public static string FromModifiedUTF7String(string str)
    {
      // RFC 3501 INTERNET MESSAGE ACCESS PROTOCOL - VERSION 4rev1
      // 5.1.3. Mailbox International Naming Convention
      // http://tools.ietf.org/html/rfc3501#section-5.1.3
      if (!str.Contains("&"))
        return str;

      var bytes = USAscii.GetBytes(str);
      var decoded = new StringBuilder();

      for (var index = 0; index < bytes.Length; index++) {
        // In modified UTF-7, printable US-ASCII characters, except for "&",
        // represent themselves
        // "&" is used to shift to modified BASE64
        if (bytes[index] != 0x26) { // '&'
          decoded.Append((char)bytes[index]);
          continue;
        }

        if (bytes.Length <= ++index)
          // incorrect form
          throw new FormatException("incorrect form");

        if (bytes[index] == 0x2d) { // '-'
          // The character "&" (0x26) is represented by the two-octet sequence "&-".
          decoded.Append('&');
          continue;
        }

        var nonprintable = new StringBuilder();

        for (; index < bytes.Length; index++) {
          if (bytes[index] == 0x2d) // '-'
            // "-" is used to shift back to US-ASCII
            break;

          nonprintable.Append((char)bytes[index]);
        }

        // modified UTF7 -> string
        decoded.Append(FromModifiedUTF7(nonprintable.ToString()));
      }

      return decoded.ToString();
    }

    private static string FromModifiedUTF7(string str)
    {
      return Encoding.BigEndianUnicode.GetString(FromModifiedBase64(str));
    }

    private static byte[] FromModifiedBase64(string str)
    {
      // "," is used instead of "/"
      str = str.Replace(',', '/');

      var padding = 4 - str.Length & 3;

      if (padding == 4)
        return Convert.FromBase64String(str);
      else if (padding == 3)
        throw new FormatException("incorrect form");
      else
        return Convert.FromBase64String(str + (new string('=', padding)));
    }

    public static string ToModifiedUTF7String(string str)
    {
      // RFC 3501 INTERNET MESSAGE ACCESS PROTOCOL - VERSION 4rev1
      // 5.1.3. Mailbox International Naming Convention
      // http://tools.ietf.org/html/rfc3501#section-5.1.3

      var encoded = new StringBuilder();
      var shiftFrom = -1;

      for (var index = 0; index < str.Length; index++) {
        var c = str[index];

        if ((0x20 <= c && c <= 0x7e)) {
          if (0 <= shiftFrom) {
            // string -> modified UTF7
            encoded.Append('&');
            encoded.Append(ToModifiedUTF7(str.Substring(shiftFrom, index - shiftFrom)));
            encoded.Append('-');

            shiftFrom = -1;
          }

          // printable US-ASCII characters
          if (c == 0x26)
            // except for "&"
            encoded.Append("&-");
          else
            encoded.Append(c);
        }
        else {
          if (shiftFrom == -1)
            shiftFrom = index;
        }
      }

      if (0 <= shiftFrom) {
        // string -> modified UTF7
        encoded.Append('&');
        encoded.Append(ToModifiedUTF7(str.Substring(shiftFrom)));
        encoded.Append('-');
      }

      return encoded.ToString();
    }

    private static string ToModifiedUTF7(string str)
    {
      return ToModifiedBase64(Encoding.BigEndianUnicode.GetBytes(str));
    }

    private static string ToModifiedBase64(byte[] bytes)
    {
      var base64 = Convert.ToBase64String(bytes).Replace('/', ',');
      var padding = base64.IndexOf('=');

      if (0 <= padding)
        return base64.Substring(0, padding);
      else
        return base64;
    }
  }
}
