using System;

using Smdn.Protocols.Imap4.Protocol;

namespace Smdn.Protocols.Imap4 {
  // DataItem types:
  //   ImapDataItem
  //     => handles 'data item names'
  //   * ImapFetchDataItem
  //       => handles 'message data item names or macro'
  //     ImapSearchCriteria
  //       => handles 'searching criteria'
  //     ImapStatusDataItem
  //       => handles 'status data item names'
  //     ImapStoreDataItem
  //       => handles 'message data item name' and 'value for message data item'

  // 6.4.5. FETCH Command
  //   message data item names or macro
  public class ImapFetchDataItem : ImapDataItem {
    // ALL
    //    Macro equivalent to: (FLAGS INTERNALDATE RFC822.SIZE ENVELOPE)
    public static readonly ImapFetchDataItem All
      = new ImapFetchDataItem("ALL", true, "FLAGS", "INTERNALDATE", "RFC822.SIZE", "ENVELOPE");

    // FAST
    //    Macro equivalent to: (FLAGS INTERNALDATE RFC822.SIZE)
    public static readonly ImapFetchDataItem Fast
      = new ImapFetchDataItem("FAST", true, "FLAGS", "INTERNALDATE", "RFC822.SIZE");

    // FULL
    //    Macro equivalent to: (FLAGS INTERNALDATE RFC822.SIZE ENVELOPE
    //    BODY)
    public static readonly ImapFetchDataItem Full
      = new ImapFetchDataItem("FULL", true, "FLAGS", "INTERNALDATE", "RFC822.SIZE", "ENVELOPE");

    // BODY
    //     Non-extensible form of BODYSTRUCTURE.
    public static readonly ImapFetchDataItem Body
      = new ImapFetchDataItem("BODY");

    // BODY[<section>]<<partial>>
    //     The text of a particular body section.  The section
    //     specification is a set of zero or more part specifiers
    //     delimited by periods.  A part specifier is either a part number
    //     or one of the following: HEADER, HEADER.FIELDS,
    //     HEADER.FIELDS.NOT, MIME, and TEXT.  An empty section
    //     specification refers to the entire message, including the
    //     header.
    public static ImapFetchDataItem BodyText()
    {
      return BodyCommon(false, null, null, null);
    }

    public static ImapFetchDataItem BodyText(string section)
    {
      return BodyCommon(false, section, null, null);
    }

    public static ImapFetchDataItem BodyText(int start)
    {
      return BodyCommon(false, null, start, null);
    }

    public static ImapFetchDataItem BodyText(int start, int count)
    {
      return BodyCommon(false, null, start, count);
    }

    public static ImapFetchDataItem BodyText(string section, int start)
    {
      return BodyCommon(false, section, start, null);
    }

    public static ImapFetchDataItem BodyText(string section, int start, int count)
    {
      return BodyCommon(false, section, start, count);
    }

    // BODY.PEEK[<section>]<<partial>>
    //     An alternate form of BODY[<section>] that does not implicitly
    //     set the \Seen flag.
    public static ImapFetchDataItem BodyPeek()
    {
      return BodyCommon(true, null, null, null);
    }

    public static ImapFetchDataItem BodyPeek(string section)
    {
      return BodyCommon(true, section, null, null);
    }

    public static ImapFetchDataItem BodyPeek(int start)
    {
      return BodyCommon(true, null, start, null);
    }

    public static ImapFetchDataItem BodyPeek(int start, int count)
    {
      return BodyCommon(true, null, start, count);
    }

    public static ImapFetchDataItem BodyPeek(string section, int start)
    {
      return BodyCommon(true, section, start, null);
    }

    public static ImapFetchDataItem BodyPeek(string section, int start, int count)
    {
      return BodyCommon(true, section, start, count);
    }

    // TODO: HEADER, HEADER.FIELDS, HEADER.FIELDS.NOT, MIME, and TEXT
    private static ImapFetchDataItem BodyCommon(bool peek, string section, int? start, int? count)
    {
      var body = peek ? "BODY.PEEK" : "BODY";

      if (start == null)
        return new ImapFetchDataItem(string.Format("{0}[{1}]",
                                                   body,
                                                   section));
      else
        if (count == null)
          return new ImapFetchDataItem(string.Format("{0}[{1}]<{2}>",
                                                     body,
                                                     section,
                                                     start.Value));
        else
          return new ImapFetchDataItem(string.Format("{0}[{1}]<{2}.{3}>",
                                                     body,
                                                     section,
                                                     start.Value,
                                                     count.Value));
    }

    // BODYSTRUCTURE
    //     The [MIME-IMB] body structure of the message.  This is computed
    //     by the server by parsing the [MIME-IMB] header fields in the
    //     [RFC-2822] header and [MIME-IMB] headers.
    public static readonly ImapFetchDataItem BodyStructure
      = new ImapFetchDataItem("BODYSTRUCTURE");

    // ENVELOPE
    //     The envelope structure of the message.  This is computed by the
    //     server by parsing the [RFC-2822] header into the component
    //     parts, defaulting various fields as necessary.
    public static readonly ImapFetchDataItem Envelope
      = new ImapFetchDataItem("ENVELOPE");

    // FLAGS
    //     The flags that are set for this message.
    public static readonly ImapFetchDataItem Flags
      = new ImapFetchDataItem("FLAGS");

    // INTERNALDATE
    //     The internal date of the message.
    public static readonly ImapFetchDataItem InternalDate
      = new ImapFetchDataItem("INTERNALDATE");

    // RFC822
    //     Functionally equivalent to BODY[], differing in the syntax of
    //     the resulting untagged FETCH data (RFC822 is returned).
    public static readonly ImapFetchDataItem Rfc822
      = new ImapFetchDataItem("RFC822");

    // RFC822.HEADER
    //     Functionally equivalent to BODY.PEEK[HEADER], differing in the
    //     syntax of the resulting untagged FETCH data (RFC822.HEADER is
    //     returned).
    public static readonly ImapFetchDataItem Rfc822Header
      = new ImapFetchDataItem("RFC822.HEADER");

    // RFC822.SIZE
    //     The [RFC-2822] size of the message.
    public static readonly ImapFetchDataItem Rfc822Size
      = new ImapFetchDataItem("RFC822.SIZE");

    // RFC822.TEXT
    //     Functionally equivalent to BODY[TEXT], differing in the syntax
    //     of the resulting untagged FETCH data (RFC822.TEXT is returned).
    public static readonly ImapFetchDataItem Rfc822Text
      = new ImapFetchDataItem("RFC822.TEXT");

    // UID
    //     The unique identifier for the message.
    public static readonly ImapFetchDataItem Uid
      = new ImapFetchDataItem("UID");

    private ImapFetchDataItem(ImapString macro, bool isMacro, params ImapString[] equivalentDataItemNames)
      : base(macro)
    {
      this.equivalent = equivalentDataItemNames;
    }

    private ImapFetchDataItem(params ImapString[] dataItemNames)
      : base(dataItemNames)
    {
      this.equivalent = null;
    }

    public ImapFetchDataItem CombineWith(ImapFetchDataItem dataItem)
    {
      return this + dataItem;
    }

    public static ImapFetchDataItem Combine(ImapFetchDataItem x, ImapFetchDataItem y)
    {
      return x + y;
    }

    public static ImapFetchDataItem operator+ (ImapFetchDataItem x, ImapFetchDataItem y)
    {
      var xx = (x.equivalent == null) ? x.DataItems : x.equivalent;
      var yy = (y.equivalent == null) ? y.DataItems : y.equivalent;

      return new ImapFetchDataItem(GetCombinedDataItems(xx, yy));
    }

    internal override ImapStringList ToStringList()
    {
      if (equivalent == null)
        // message data item names
        return new ImapParenthesizedString(base.DataItems);
      else
        // macro
        return base.ToStringList();
    }

    private readonly ImapString[] equivalent;
  }
}
