using System;
using System.Collections.Generic;

namespace Smdn.Protocols.Imap4 {
  /// <remarks>
  /// ImapMailboxのインスタンスはImapMailboxManagerで管理する
  /// ImapMailboxManagerで保持していないImapMailboxが見付かった場合は、
  /// このクラスでインスタンスを作成してImapMailboxManagerに追加する.
  /// 名前、IMAP URLでのメールボックスの参照を行う場合は、このクラスで
  /// 管理しているインスタンスを返す
  /// </remarks>
  // TODO: 名前空間の考慮
  internal class ImapMailboxManager {
    internal static ImapMailbox GetMailbox(ImapServerInfo serverInfo, string name)
    {
      var mailboxes = GetMailboxes(serverInfo, true);

      if (name.Equals("INBOX", StringComparison.InvariantCultureIgnoreCase))
        name = "INBOX";

      if (mailboxes.ContainsKey(name)) {
        return mailboxes[name];
      }
      else {
        var mailbox = new ImapMailbox(serverInfo, name);

        lock (mailboxes) {
          mailboxes.Add(mailbox.Name, mailbox);
        }

        return mailbox;
      }
    }

    internal static ImapMailbox GetExistMailbox(ImapServerInfo serverInfo, ImapUri mailboxUri)
    {
      if (serverInfo.Uri.UserAtHost != mailboxUri.UserAtHost)
        throw new InvalidOperationException("requested URI is not in that server");

      var mailboxName = mailboxUri.Mailbox;

      if (mailboxName.Contains("/")) {
        // get hierarchy delimiter of top level mailbox
        var topMailboxUri = mailboxUri;
        ImapMailbox topMailbox = null;

        for (;;) {
          topMailboxUri = topMailboxUri.GetParent();

          if (topMailboxUri == null)
            break;

          if (topMailboxUri.Mailbox.Contains("/"))
            continue;

          topMailbox = GetExistMailbox(serverInfo, topMailboxUri);

          break;
        }

        if (topMailbox != null && !string.IsNullOrEmpty(topMailbox.HierarchyDelimiter))
          // replace delimiter
          mailboxName = mailboxName.Replace("/", topMailbox.HierarchyDelimiter);
      }

      if (mailboxName.Equals("INBOX", StringComparison.InvariantCultureIgnoreCase))
        mailboxName = "INBOX";

      return GetExistMailbox(serverInfo, mailboxName);
    }

    internal static ImapMailbox GetExistMailbox(ImapServerInfo serverInfo, string mailboxName)
    {
      var mailboxes = GetMailboxes(serverInfo, true);

      if (mailboxes.ContainsKey(mailboxName))
        return mailboxes[mailboxName];
      else
        return null;
    }

    internal static ImapMailbox[] GetExistMailboxes(ImapServerInfo serverInfo)
    {
      var dict = GetMailboxes(serverInfo, false);

      if (dict == null)
        return null;

      var mailboxes = new List<ImapMailbox>(dict.Values); 

      return mailboxes.ToArray();
    }

    internal static void DeleteMailbox(ImapServerInfo serverInfo, ImapMailbox mailbox)
    {
      DeleteMailbox(serverInfo, mailbox.Name);
    }

    internal static void DeleteMailbox(ImapServerInfo serverInfo, string mailboxName)
    {
      if (serverMailboxes[serverInfo].ContainsKey(mailboxName))
        serverMailboxes[serverInfo].Remove(mailboxName);
    }

    internal static ImapMailbox GetRenamedMailbox(ImapServerInfo serverInfo, string existingMailboxName, string newMailboxName)
    {
      var mailbox = GetExistMailbox(serverInfo, existingMailboxName);

      if (mailbox != null && !mailbox.Name.Equals("INBOX", StringComparison.InvariantCultureIgnoreCase))
        DeleteMailbox(serverInfo, mailbox);

      return GetMailbox(serverInfo, newMailboxName);
    }

    internal static ImapUri GetMailboxUrl(ImapMailbox mailbox)
    {
      var name = mailbox.Name;

      if (!string.IsNullOrEmpty(mailbox.HierarchyDelimiter))
        name = name.Replace(mailbox.HierarchyDelimiter, "/");

      return new ImapUri(mailbox.ServerInfo.Uri, name); 
    }

    private static Dictionary<string, ImapMailbox> GetMailboxes(ImapServerInfo serverInfo, bool createIfNonExistent)
    {
      if (!serverMailboxes.ContainsKey(serverInfo)) {
        if (createIfNonExistent) {
          lock (serverMailboxes) {
            serverMailboxes.Add(serverInfo, new Dictionary<string, ImapMailbox>());
          }
        }
        else {
          return null;
        }
      }

      return serverMailboxes[serverInfo];
    }

    private static Dictionary<ImapServerInfo, Dictionary<string, ImapMailbox>> serverMailboxes =
      new Dictionary<ImapServerInfo, Dictionary<string, ImapMailbox>>();
  }
}
