using System;
using System.Text;
using System.Web;

using Smdn.Protocols.Imap4.Protocol;

namespace Smdn.Protocols.Imap4 {
  // DataItem types:
  //   ImapDataItem
  //     => handles 'data item names'
  //     ImapFetchDataItem
  //       => handles 'message data item names or macro'
  //   * ImapSearchCriteria
  //       => handles 'searching criteria'
  //     ImapStatusDataItem
  //       => handles 'status data item names'
  //     ImapStoreDataItem
  //       => handles 'message data item name' and 'value for message data item'

  // 6.4.4. SEARCH Command
  //   searching criteria
  public class ImapSearchCriteria : ImapDataItem {
    //  <sequence set>
    //     Messages with message sequence numbers corresponding to the
    //     specified message sequence number set.
    public static ImapSearchCriteria SequenceSet(ImapSequenceSet sequenceSet)
    {
      return new ImapSearchCriteria(sequenceSet.ToString());
    }

    // ALL
    //     All messages in the mailbox; the default initial key for
    //     ANDing.
    public static readonly ImapSearchCriteria All
      = new ImapSearchCriteria("ALL");

    // ANSWERED
    //     Messages with the \Answered flag set.
    public static readonly ImapSearchCriteria Answered
      = new ImapSearchCriteria("ANSWERED");

    // BCC <string>
    //     Messages that contain the specified string in the envelope
    //     structure's BCC field.
    public static ImapSearchCriteria Bcc(string str)
    {
      return new ImapSearchCriteria("BCC", new ImapQuotedString(str));
    }

    // BEFORE <date>
    //     Messages whose internal date (disregarding time and timezone)
    //     is earlier than the specified date.
    public static ImapSearchCriteria Before(DateTime date)
    {
      return new ImapSearchCriteria("BEFORE", ImapConverter.ToDateString(date));
    }

    // BODY <string>
    //     Messages that contain the specified string in the body of the
    //     message.
    public static ImapSearchCriteria Body(string str)
    {
      return new ImapSearchCriteria("BODY", new ImapLiteralString(str));
    }

    // CC <string>
    //     Messages that contain the specified string in the envelope
    //     structure's CC field.
    public static ImapSearchCriteria Cc(string str)
    {
      return new ImapSearchCriteria("CC", new ImapQuotedString(str));
    }

    // DELETED
    //     Messages with the \Deleted flag set.
    public static readonly ImapSearchCriteria Deleted
      = new ImapSearchCriteria("DELETED");

    // DRAFT
    //     Messages with the \Draft flag set.
    public static readonly ImapSearchCriteria Draft
      = new ImapSearchCriteria("DRAFT");

    // FLAGGED
    //     Messages with the \Flagged flag set.
    public static readonly ImapSearchCriteria Flagged
      = new ImapSearchCriteria("FLAGGED");

    // FROM <string>
    //     Messages that contain the specified string in the envelope
    //     structure's FROM field.
    public static ImapSearchCriteria From(string str)
    {
      return new ImapSearchCriteria("FROM", new ImapQuotedString(str));
    }

    // HEADER <field-name> <string>
    //     Messages that have a header with the specified field-name (as
    //     defined in [RFC-2822]) and that contains the specified string
    //     in the text of the header (what comes after the colon).  If the
    //     string to search is zero-length, this matches all messages that
    //     have a header line with the specified field-name regardless of
    //     the contents.
    public static ImapSearchCriteria Header(string fieldName, string str)
    {
      return new ImapSearchCriteria("HEADER", fieldName, new ImapQuotedString(str));
    }

    // KEYWORD <flag>
    //     Messages with the specified keyword flag set.
    public static ImapSearchCriteria Keyword(string keyword)
    {
      return new ImapSearchCriteria("KEYWORD", keyword);
    }

    // LARGER <n>
    //     Messages with an [RFC-2822] size larger than the specified
    //     number of octets.
    public static ImapSearchCriteria Larger(int size)
    {
      if (size < 0)
        throw new ArgumentException("size must be greater than 0");

      return new ImapSearchCriteria("LARGER", size.ToString());
    }

    // NEW
    //     Messages that have the \Recent flag set but not the \Seen flag.
    //     This is functionally equivalent to "(RECENT UNSEEN)".
    public static readonly ImapSearchCriteria New
      = new ImapSearchCriteria("NEW");

    // OLD
    //     Messages that do not have the \Recent flag set.  This is
    //     functionally equivalent to "NOT RECENT" (as opposed to "NOT
    //     NEW").
    public static readonly ImapSearchCriteria Old
      = new ImapSearchCriteria("OLD");

    // ON <date>
    //     Messages whose internal date (disregarding time and timezone)
    //     is within the specified date.
    public static ImapSearchCriteria On(DateTime date)
    {
      return new ImapSearchCriteria("ON", ImapConverter.ToDateString(date));
    }

    // RECENT
    //     Messages that have the \Recent flag set.
    public static readonly ImapSearchCriteria Recent
      = new ImapSearchCriteria("RECENT");

    // SEEN
    //     Messages that have the \Seen flag set.
    public static readonly ImapSearchCriteria Seen
      = new ImapSearchCriteria("SEEN");

    // SENTBEFORE <date>
    //     Messages whose [RFC-2822] Date: header (disregarding time and
    //     timezone) is earlier than the specified date.
    public static ImapSearchCriteria SentBefore(DateTime date)
    {
      return new ImapSearchCriteria("SENTBEFORE", ImapConverter.ToDateString(date));
    }

    // SENTON <date>
    //     Messages whose [RFC-2822] Date: header (disregarding time and
    //     timezone) is within the specified date.
    public static ImapSearchCriteria SentOn(DateTime date)
    {
      return new ImapSearchCriteria("SENTON", ImapConverter.ToDateString(date));
    }

    // SENTSINCE <date>
    //     Messages whose [RFC-2822] Date: header (disregarding time and
    //     timezone) is within or later than the specified date.
    public static ImapSearchCriteria SentSince(DateTime date)
    {
      return new ImapSearchCriteria("SENTSINCE", ImapConverter.ToDateString(date));
    }

    // SINCE <date>
    //     Messages whose internal date (disregarding time and timezone)
    //     is within or later than the specified date.
    public static ImapSearchCriteria Since(DateTime date)
    {
      return new ImapSearchCriteria("SINCE", ImapConverter.ToDateString(date));
    }

    // SMALLER <n>
    //     Messages with an [RFC-2822] size smaller than the specified
    //     number of octets.
    public static ImapSearchCriteria Smaller(int size)
    {
      if (size < 0)
        throw new ArgumentException("size must be greater than 0");

      return new ImapSearchCriteria("SMALLER", size.ToString());
    }

    // SUBJECT <string>
    //     Messages that contain the specified string in the envelope
    //     structure's SUBJECT field.
    public static ImapSearchCriteria Subject(string str)
    {
      return new ImapSearchCriteria("SUBJECT", new ImapLiteralString(str));
    }

    // TEXT <string>
    //     Messages that contain the specified string in the header or
    //     body of the message.
    public static ImapSearchCriteria Text(string str)
    {
      return new ImapSearchCriteria("TEXT", new ImapLiteralString(str));
    }

    // TO <string>
    //     Messages that contain the specified string in the envelope
    //     structure's TO field.
    public static ImapSearchCriteria To(string str)
    {
      return new ImapSearchCriteria("TO", new ImapQuotedString(str));
    }

    // UID <sequence set>
    //     Messages with unique identifiers corresponding to the specified
    //     unique identifier set.  Sequence set ranges are permitted.
    public static ImapSearchCriteria Uid(ImapSequenceSet sequenceSet)
    {
      return new ImapSearchCriteria("UID", sequenceSet.ToString());
    }

    // UNANSWERED
    //     Messages that do not have the \Answered flag set.
    public static readonly ImapSearchCriteria Unanswered
      = new ImapSearchCriteria("UNANSWERED");

    // UNDELETED
    //     Messages that do not have the \Deleted flag set.
    public static readonly ImapSearchCriteria Undeleted
      = new ImapSearchCriteria("UNDELETED");

    // UNDRAFT
    //     Messages that do not have the \Draft flag set.
    public static readonly ImapSearchCriteria Undraft
      = new ImapSearchCriteria("UNDRAFT");

    // UNFLAGGED
    //     Messages that do not have the \Flagged flag set.
    public static readonly ImapSearchCriteria Unflagged
      = new ImapSearchCriteria("UNFLAGGED");

    // UNKEYWORD <flag>
    //     Messages that do not have the specified keyword flag set.
    public static ImapSearchCriteria Unkeyword(string keyword)
    {
      return new ImapSearchCriteria("UNKEYWORD", keyword);
    }

    // UNSEEN
    //     Messages that do not have the \Seen flag set.
    public static readonly ImapSearchCriteria Unseen
      = new ImapSearchCriteria("UNSEEN");

    // NOT <search-key>
    //     Messages that do not match the specified search key.
    public ImapSearchCriteria Not()
    {
      return !this;
    }

    public static ImapSearchCriteria Not(ImapSearchCriteria c)
    {
      return !c;
    }

    public static ImapSearchCriteria operator! (ImapSearchCriteria c)
    {
      return new ImapSearchCriteria("NOT", new ImapParenthesizedString(c.DataItems));
    }

    // OR <search-key1> <search-key2>
    //     Messages that match either search key.
    public ImapSearchCriteria Or(ImapSearchCriteria searchingCriteria)
    {
      return this | searchingCriteria;
    }

    public static ImapSearchCriteria Or(ImapSearchCriteria x, ImapSearchCriteria y)
    {
      return x | y;
    }

    public static ImapSearchCriteria operator| (ImapSearchCriteria x, ImapSearchCriteria y)
    {
      return new ImapSearchCriteria("OR", new ImapParenthesizedString(x.DataItems), new ImapParenthesizedString(y.DataItems));
    }

    // AND
    public ImapSearchCriteria And(ImapSearchCriteria searchingCriteria)
    {
      return this & searchingCriteria;
    }

    public static ImapSearchCriteria And(ImapSearchCriteria x, ImapSearchCriteria y)
    {
      return x & y;
    }

    public static ImapSearchCriteria operator& (ImapSearchCriteria x, ImapSearchCriteria y)
    {
      return new ImapSearchCriteria(GetCombinedDataItems(x, y));
    }

    internal static ImapSearchCriteria FromUri(ImapUri uri, Encoding encoding)
    {
      if (uri.Query.Length <= 1)
        return new ImapSearchCriteria(string.Empty);
      else
        return new ImapSearchCriteria(HttpUtility.UrlDecode(uri.Query.Substring(1), encoding));
    }

    private ImapSearchCriteria(params ImapString[] dataItem)
      : base(dataItem)
    {
    }
  }
}