using System;

using Smdn.Protocols.Imap4.Protocol;

namespace Smdn.Protocols.Imap4 {
  // DataItem types:
  //   ImapDataItem
  //     => handles 'data item names'
  //     ImapFetchDataItem
  //       => handles 'message data item names or macro'
  //     ImapSearchCriteria
  //       => handles 'searching criteria'
  //     ImapStatusDataItem
  //       => handles 'status data item names'
  //   * ImapStoreDataItem
  //       => handles 'message data item name' and 'value for message data item'

  // 6.4.6. STORE Command
  //   message data item name
  //   value for message data item
  public class ImapStoreDataItem : ImapDataItem {

    // FLAGS <flag list>
    //     Replace the flags for the message (other than \Recent) with the
    //     argument.  The new value of the flags is returned as if a FETCH
    //     of those flags was done.
    public static ImapStoreDataItem ReplaceFlags(ImapSystemFlags systemFlags)
    {
      return ReplaceFlags(systemFlags, new string[] {});
    }

    public static ImapStoreDataItem ReplaceFlags(params string[] keywords)
    {
      return ReplaceFlags(ImapSystemFlags.None, keywords);
    }

    public static ImapStoreDataItem ReplaceFlags(ImapMessageFlag flags)
    {
      return ReplaceFlags(flags.SystemFlags, flags.Keywords);
    }

    public static ImapStoreDataItem ReplaceFlags(ImapSystemFlags systemFlags, params string[] keywords)
    {
      return FlagsCommon("FLAGS", systemFlags, keywords);
    }

    // FLAGS.SILENT <flag list>
    //     Equivalent to FLAGS, but without returning a new value.
    public static ImapStoreDataItem ReplaceFlagsSilent(ImapSystemFlags systemFlags)
    {
      return ReplaceFlagsSilent(systemFlags, new string[] {});
    }

    public static ImapStoreDataItem ReplaceFlagsSilent(params string[] keywords)
    {
      return ReplaceFlagsSilent(ImapSystemFlags.None, keywords);
    }

    public static ImapStoreDataItem ReplaceFlagsSilent(ImapMessageFlag flags)
    {
      return ReplaceFlagsSilent(flags.SystemFlags, flags.Keywords);
    }

    public static ImapStoreDataItem ReplaceFlagsSilent(ImapSystemFlags systemFlags, params string[] keywords)
    {
      return FlagsCommon("FLAGS.SILENT", systemFlags, keywords);
    }

    // +FLAGS <flag list>
    //     Add the argument to the flags for the message.  The new value
    //     of the flags is returned as if a FETCH of those flags was done.
    public static ImapStoreDataItem AddFlags(ImapSystemFlags systemFlags)
    {
      return AddFlags(systemFlags, new string[] {});
    }

    public static ImapStoreDataItem AddFlags(params string[] keywords)
    {
      return AddFlags(ImapSystemFlags.None, keywords);
    }

    public static ImapStoreDataItem AddFlags(ImapMessageFlag flags)
    {
      return AddFlags(flags.SystemFlags, flags.Keywords);
    }

    public static ImapStoreDataItem AddFlags(ImapSystemFlags systemFlags, params string[] keywords)
    {
      return FlagsCommon("+FLAGS", systemFlags, keywords);
    }

    // +FLAGS.SILENT <flag list>
    //     Equivalent to +FLAGS, but without returning a new value.
    public static ImapStoreDataItem AddFlagsSilent(ImapSystemFlags systemFlags)
    {
      return AddFlagsSilent(systemFlags, new string[] {});
    }

    public static ImapStoreDataItem AddFlagsSilent(params string[] keywords)
    {
      return AddFlagsSilent(ImapSystemFlags.None, keywords);
    }

    public static ImapStoreDataItem AddFlagsSilent(ImapMessageFlag flags)
    {
      return AddFlagsSilent(flags.SystemFlags, flags.Keywords);
    }

    public static ImapStoreDataItem AddFlagsSilent(ImapSystemFlags systemFlags, params string[] keywords)
    {
      return FlagsCommon("+FLAGS.SILENT", systemFlags, keywords);
    }

    // -FLAGS <flag list>
    //     Remove the argument from the flags for the message.  The new
    //     value of the flags is returned as if a FETCH of those flags was
    //     done.
    public static ImapStoreDataItem RemoveFlags(ImapSystemFlags systemFlags)
    {
      return RemoveFlags(systemFlags, new string[] {});
    }

    public static ImapStoreDataItem RemoveFlags(params string[] keywords)
    {
      return RemoveFlags(ImapSystemFlags.None, keywords);
    }

    public static ImapStoreDataItem RemoveFlags(ImapMessageFlag flags)
    {
      return RemoveFlags(flags.SystemFlags, flags.Keywords);
    }

    public static ImapStoreDataItem RemoveFlags(ImapSystemFlags systemFlags, params string[] keywords)
    {
      return FlagsCommon("-FLAGS", systemFlags, keywords);
    }

    // -FLAGS.SILENT <flag list>
    //     Equivalent to -FLAGS, but without returning a new value.
    public static ImapStoreDataItem RemoveFlagsSilent(ImapSystemFlags systemFlags)
    {
      return RemoveFlagsSilent(systemFlags, new string[] {});
    }

    public static ImapStoreDataItem RemoveFlagsSilent(params string[] keywords)
    {
      return RemoveFlagsSilent(ImapSystemFlags.None, keywords);
    }

    public static ImapStoreDataItem RemoveFlagsSilent(ImapMessageFlag flags)
    {
      return RemoveFlagsSilent(flags.SystemFlags, flags.Keywords);
    }

    public static ImapStoreDataItem RemoveFlagsSilent(ImapSystemFlags systemFlags, params string[] keywords)
    {
      return FlagsCommon("-FLAGS.SILENT", systemFlags, keywords);
    }

    private static ImapStoreDataItem FlagsCommon(ImapString dataItem, ImapSystemFlags systemFlags, string[] keywords)
    {
      // value for message data item(flag list)
      var flagList = ImapConverter.ToFlagList(true, systemFlags, keywords);

      return new ImapStoreDataItem(dataItem, flagList);
    }

    private ImapStoreDataItem(params ImapString[] dataItems)
      : base(dataItems)
    {
    }
  }
}