using System;
using System.Collections.Generic;

namespace Smdn.Protocols.Imap4.Protocol {
  // format converting methods for ImapStatusResponse
  public static class ImapStatusResponseConverter {
    /*
     * 7.1. Server Responses - Status Responses
     */

    //      ALERT
    //         The human-readable text contains a special alert that MUST be
    //         presented to the user in a fashion that calls the user's
    //         attention to the message.
    public static bool FromAlert(ImapStatusResponse status, out string message)
    {
      message = null;

      if (!IsRequiredStatusResponse(status, "ALERT"))
        return false;

      message = status.Text;

      return true;
    }

    //      BADCHARSET
    //         Optionally followed by a parenthesized list of charsets.  A
    //         SEARCH failed because the given charset is not supported by
    //         this implementation.  If the optional list of charsets is
    //         given, this lists the charsets that are supported by this
    //         implementation.
    public static bool FromBadCharset(ImapStatusResponse status, out string[] supportedCharsets)
    {
      supportedCharsets = null;

      if (!IsRequiredStatusResponse(status, "BADCHARSET"))
        return false;

      var list = ImapDataConverter.ToTextList(status.ResponseCode.Data);

      if (list == null) {
        return false;
      }
      else {
        supportedCharsets = list;
        return true;
      }
    }

    //      CAPABILITY
    //         Followed by a list of capabilities.  This can appear in the
    //         initial OK or PREAUTH response to transmit an initial
    //         capabilities list.  This makes it unnecessary for a client to
    //         send a separate CAPABILITY command if it recognizes this
    //         response.
    public static bool FromCapability(ImapStatusResponse status, out ImapCapability capabilities)
    {
      capabilities = null;

      if (!IsRequiredStatusResponse(status, "CAPABILITY"))
        return false;

      // insert status text 'CAPABILITY' to response data list
      var list = new List<ImapData>(status.ResponseCode.Data);

      list.Insert(0, ImapData.CreateTextData("CAPABILITY"));

      return ImapDataConverter.ToCapabilityData(list.ToArray(), out capabilities);
    }

    //      PARSE
    //         The human-readable text represents an error in parsing the
    //         [RFC-2822] header or [MIME-IMB] headers of a message in the
    //         mailbox.
    public static bool FromParse(ImapStatusResponse status, out string message)
    {
      message = null;

      if (!IsRequiredStatusResponse(status, "PARSE"))
        return false;

      message = status.Text;

      return true;
    }

    //      PERMANENTFLAGS
    //         Followed by a parenthesized list of flags, indicates which of
    //         the known flags the client can change permanently.  Any flags
    //         that are in the FLAGS untagged response, but not the
    //         PERMANENTFLAGS list, can not be set permanently.  If the client
    //         attempts to STORE a flag that is not in the PERMANENTFLAGS
    //         list, the server will either ignore the change or store the
    //         state change for the remainder of the current session only.
    //         The PERMANENTFLAGS list can also include the special flag \*,
    //         which indicates that it is possible to create new keywords by
    //         attempting to store those flags in the mailbox.
    public static bool FromPermanentFlags(ImapStatusResponse status, out ImapMessageFlag flags)
    {
      flags = null;

      if (!IsRequiredStatusResponse(status, "PERMANENTFLAGS"))
        return false;

      if (status.ResponseCode.Data.Length < 1)
        return false;

      return ImapDataConverter.ToFlag(status.ResponseCode.Data[0].List, out flags);
    }

    //      READ-ONLY
    //         The mailbox is selected read-only, or its access while selected
    //         has changed from read-write to read-only.
    public static bool FromReadOnly(ImapStatusResponse status)
    {
      return IsRequiredStatusResponse(status, "READ-ONLY");
    }

    //      READ-WRITE
    //         The mailbox is selected read-write, or its access while
    //         selected has changed from read-only to read-write.
    public static bool FromReadWrite(ImapStatusResponse status)
    {
      return IsRequiredStatusResponse(status, "READ-WRITE");
    }

    //      TRYCREATE
    //         An APPEND or COPY attempt is failing because the target mailbox
    //         does not exist (as opposed to some other reason).  This is a
    //         hint to the client that the operation can succeed if the
    //         mailbox is first created by the CREATE command.
    public static bool FromTryCreate(ImapStatusResponse status)
    {
      return IsRequiredStatusResponse(status, "TRYCREATE");
    }

    //      UIDNEXT
    //         Followed by a decimal number, indicates the next unique
    //         identifier value.  Refer to section 2.3.1.1 for more
    //         information.
    public static bool FromUidNext(ImapStatusResponse status, out int uidNext)
    {
      uidNext = 0;

      if (!IsRequiredStatusResponse(status, "UIDNEXT"))
        return false;

      if (status.ResponseCode.Data.Length < 1)
        return false;

      var val = ImapDataConverter.ToNumber(status.ResponseCode.Data[0]);

      if (val == null) {
        return false;
      }
      else {
        uidNext = (int)val;
        return true;
      }
    }

    //      UIDVALIDITY
    //         Followed by a decimal number, indicates the unique identifier
    //         validity value.  Refer to section 2.3.1.1 for more information.
    public static bool FromUidValidity(ImapStatusResponse status, out int uidValidity)
    {
      uidValidity = 0;

      if (!IsRequiredStatusResponse(status, "UIDVALIDITY"))
        return false;

      if (status.ResponseCode.Data.Length < 1)
        return false;

      var val = ImapDataConverter.ToNumber(status.ResponseCode.Data[0]);

      if (val == null) {
        return false;
      }
      else {
        uidValidity = (int)val;
        return true;
      }
    }

    //      UNSEEN
    //         Followed by a decimal number, indicates the number of the first
    //         message without the \Seen flag set.
    public static bool FromUnseen(ImapStatusResponse status, out int unseen)
    {
      unseen = 0;

      if (!IsRequiredStatusResponse(status, "UNSEEN"))
        return false;

      if (status.ResponseCode.Data.Length < 1)
        return false;

      var val = ImapDataConverter.ToNumber(status.ResponseCode.Data[0]);
    
      if (val == null) {
        return false;
      }
      else {
        unseen = (int)val;
        return true;
      }
    }

    // RFC 2221 IMAP4 Login Referrals
    // http://tools.ietf.org/html/rfc2221
    // 4.1. LOGIN and AUTHENTICATE Referrals
    //    An IMAP4 server MAY respond to a LOGIN or AUTHENTICATE command with a
    //    home server referral if it wishes to direct the user to another IMAP4
    //    server.
    // 
    // 4.2. BYE at connection startup referral
    //    An IMAP4 server MAY respond with an untagged BYE and a REFERRAL
    //    response code that contains an IMAP URL to a home server if it is not
    //    willing to accept connections and wishes to direct the client to
    //    another IMAP4 server.
    // 
    // 5. Formal Syntax
    //    resp_text_code =/ "REFERRAL" SPACE <imapurl>
    //       ; See [IMAP-URL] for definition of <imapurl>
    //       ; See [RFC-2060] for base definition of resp_text_code
    // 
    // 
    // RFC 2193 IMAP4 Mailbox Referrals
    // http://tools.ietf.org/html/rfc2193
    // 4.1. SELECT, EXAMINE, DELETE, SUBSCRIBE, UNSUBSCRIBE, STATUS and APPEND
    //      Referrals
    //    An IMAP4 server MAY respond to the SELECT, EXAMINE, DELETE,
    //    SUBSCRIBE, UNSUBSCRIBE, STATUS or APPEND command with one or more
    //    IMAP mailbox referrals to indicate to the client that the mailbox is
    //    hosted on a remote server.
    //
    // 4.2. CREATE Referrals
    //    An IMAP4 server MAY respond to the CREATE command with one or more
    //    IMAP mailbox referrals, if it wishes to direct the client to issue
    //    the CREATE against another server.  The server can employ any means,
    //    such as examining the hierarchy of the specified mailbox name, in
    //    determining which server the mailbox should be created on.
    //
    // 4.3. RENAME Referrals
    //    An IMAP4 server MAY respond to the RENAME command with one or more
    //    pairs of IMAP mailbox referrals.  In each pair of IMAP mailbox
    //    referrals, the first one is an URL to the existing mailbox name and
    //    the second is an URL to the requested new mailbox name.
    //
    // 4.4. COPY Referrals
    //    An IMAP4 server MAY respond to the COPY command with one or more IMAP
    //    mailbox referrals.  This indicates that the destination mailbox is on
    //    a remote server.  To achieve the same behavior of a server COPY, the
    //    client MAY issue the constituent FETCH and APPEND commands against
    //    both servers.
    //
    // 6. Formal Syntax
    //    The following syntax specification uses the augmented Backus-Naur
    //    Form (BNF) as described in [ABNF].
    // 
    //    list_mailbox = <list_mailbox> as defined in [RFC-2060]
    //    mailbox = <mailbox> as defined in [RFC-2060]
    //    mailbox_referral = <tag> SPACE "NO" SPACE
    //       <referral_response_code> (text / text_mime2)
    //       ; See [RFC-2060] for <tag>, text and text_mime2 definition
    //    referral_response_code = "[" "REFERRAL" 1*(SPACE <url>) "]"
    //       ; See [RFC-1738] for <url> definition
    public static bool FromReferral(ImapStatusResponse status, out ImapUri[] referrals)
    {
      referrals = null;

      if (!IsRequiredStatusResponse(status, "REFERRAL"))
        return false;

      if (status.ResponseCode.Data.Length < 1)
        return false;

      var referralsList = new List<ImapUri>();

      foreach (var data in status.ResponseCode.Data) {
        var referral = ImapDataConverter.ToUri(data);
        if (referral != null)
          referralsList.Add(referral);
      }

      if (0 < referralsList.Count) {
        referrals = referralsList.ToArray();
        return true;
      }
      else {
        return false;
      }
    }

    /*
     * required/allowed command responses
     */
    public static bool FromGreetingResponse(ImapStatusResponse status, out ImapCapability capabilities, out ImapUri referral)
    {
      capabilities = null;
      referral = null;

      // 'This can appear in the initial OK or PREAUTH response'
      if (status.Status == ImapProtocolStatus.Ok || status.Status == ImapProtocolStatus.PreAuth)
        if (FromCapability(status, out capabilities))
          return true;

      // RFC 2221 IMAP4 Login Referrals
      // http://tools.ietf.org/html/rfc2221
      // 4.2. BYE at connection startup referral
      if (status.Status == ImapProtocolStatus.Bye) {
        ImapUri[] referrals;
        if (FromReferral(status, out referrals)) {
          referral = referrals[0];
          return true;
        }
      }

      return false;
    }

    public static bool FromLoginResponse(ImapStatusResponse status, out ImapUri referral)
    {
      referral = null;

      // RFC 2221 IMAP4 Login Referrals
      // http://tools.ietf.org/html/rfc2221
      // 4.1. LOGIN and AUTHENTICATE Referrals
      //    An IMAP4 server MAY respond to a LOGIN or AUTHENTICATE command with a
      //    home server referral if it wishes to direct the user to another IMAP4
      //    server.
      if (status.Status == ImapProtocolStatus.Ok || status.Status == ImapProtocolStatus.No) {
        ImapUri[] referrals;
        if (FromReferral(status, out referrals)) {
          referral = referrals[0];
          return true;
        }
      }

      return false;
    }

    public static bool FromAuthenticateResponse(ImapStatusResponse status, out ImapUri referral)
    {
      referral = null;

      // RFC 2221 IMAP4 Login Referrals
      // http://tools.ietf.org/html/rfc2221
      // 4.1. LOGIN and AUTHENTICATE Referrals
      //    An IMAP4 server MAY respond to a LOGIN or AUTHENTICATE command with a
      //    home server referral if it wishes to direct the user to another IMAP4
      //    server.
      if (status.Status == ImapProtocolStatus.Ok || status.Status == ImapProtocolStatus.No) {
        ImapUri[] referrals;
        if (FromReferral(status, out referrals)) {
          referral = referrals[0];
          return true;
        }
      }

      return false;
    }

    public static bool FromSearchNoResponse(ImapStatusResponse status, out string[] supportedCharset)
    {
      supportedCharset = null;

      if (status.Status != ImapProtocolStatus.No)
        return false;

      // 'A SEARCH failed because the given charset is not
      //  supported by this implementation'
      return FromBadCharset(status, out supportedCharset);
    }

    // 6.3.1. SELECT Command
    // Responses:  REQUIRED untagged responses: FLAGS, EXISTS, RECENT
    //             REQUIRED OK untagged responses:  UNSEEN,  PERMANENTFLAGS,
    //             UIDNEXT, UIDVALIDITY
    public static bool FromSelectExamineOkResponse(ImapStatusResponse status, out bool? readOnly, out int? unseen, out ImapMessageFlag permanentFlags, out int? uidNext, out int? uidValidity)
    {
      readOnly                = null;
      unseen                  = null;
      permanentFlags          = null;
      uidNext                 = null;
      uidValidity             = null;

      if (status.Status != ImapProtocolStatus.Ok)
        return false;

      // READ-ONLY
      if (FromReadOnly(status)) {
        readOnly = true;
        return true;
      }

      // READ-WRITE
      if (FromReadWrite(status)) {
        readOnly = false;
        return true;
      }

      // PERMANENTFLAGS
      ImapMessageFlag respFlags;

      if (FromPermanentFlags(status, out respFlags)) {
        permanentFlags = respFlags;
        return true;
      }

      // UIDNEXT
      int respUidNext;

      if (FromUidNext(status, out respUidNext)) {
        uidNext = respUidNext;
        return true;
      }

      // UIDVALIDITY
      int respUidValidity;

      if (FromUidValidity(status, out respUidValidity)) {
        uidValidity = respUidValidity;
        return true;
      }

      // UNSEEN
      int respUnseen;

      if (FromUnseen(status, out respUnseen)) {
        unseen = respUnseen;
        return true;
      }

      return false;
    }

    public static bool FromSelectExamineNoResponse(ImapStatusResponse status, out ImapUri[] referrals)
    {
      referrals = null;

      if (status.Status != ImapProtocolStatus.No)
        return false;

      // RFC 2193 IMAP4 Mailbox Referrals
      // http://tools.ietf.org/html/rfc2193
      // 4.1. SELECT, EXAMINE, DELETE, SUBSCRIBE, UNSUBSCRIBE, STATUS and APPEND
      //      Referrals
      //    An IMAP4 server MAY respond to the SELECT, EXAMINE, DELETE,
      //    SUBSCRIBE, UNSUBSCRIBE, STATUS or APPEND command with one or more
      //    IMAP mailbox referrals to indicate to the client that the mailbox is
      //    hosted on a remote server.
      return FromReferral(status, out referrals);
    }

    public static bool FromDeleteNoResponse(ImapStatusResponse status, out ImapUri[] referrals)
    {
      referrals = null;

      if (status.Status != ImapProtocolStatus.No)
        return false;

      // RFC 2193 IMAP4 Mailbox Referrals
      // http://tools.ietf.org/html/rfc2193
      // 4.1. SELECT, EXAMINE, DELETE, SUBSCRIBE, UNSUBSCRIBE, STATUS and APPEND
      //      Referrals
      //    An IMAP4 server MAY respond to the SELECT, EXAMINE, DELETE,
      //    SUBSCRIBE, UNSUBSCRIBE, STATUS or APPEND command with one or more
      //    IMAP mailbox referrals to indicate to the client that the mailbox is
      //    hosted on a remote server.
      return FromReferral(status, out referrals);
    }

    public static bool FromSubscribeUnsubscribeNoResponse(ImapStatusResponse status, out ImapUri[] referrals)
    {
      referrals = null;

      if (status.Status != ImapProtocolStatus.No)
        return false;

      // RFC 2193 IMAP4 Mailbox Referrals
      // http://tools.ietf.org/html/rfc2193
      // 4.1. SELECT, EXAMINE, DELETE, SUBSCRIBE, UNSUBSCRIBE, STATUS and APPEND
      //      Referrals
      //    An IMAP4 server MAY respond to the SELECT, EXAMINE, DELETE,
      //    SUBSCRIBE, UNSUBSCRIBE, STATUS or APPEND command with one or more
      //    IMAP mailbox referrals to indicate to the client that the mailbox is
      //    hosted on a remote server.
      return FromReferral(status, out referrals);
    }

    public static bool FromStatusNoResponse(ImapStatusResponse status, out ImapUri[] referrals)
    {
      referrals = null;

      if (status.Status != ImapProtocolStatus.No)
        return false;

      // RFC 2193 IMAP4 Mailbox Referrals
      // http://tools.ietf.org/html/rfc2193
      // 4.1. SELECT, EXAMINE, DELETE, SUBSCRIBE, UNSUBSCRIBE, STATUS and APPEND
      //      Referrals
      //    An IMAP4 server MAY respond to the SELECT, EXAMINE, DELETE,
      //    SUBSCRIBE, UNSUBSCRIBE, STATUS or APPEND command with one or more
      //    IMAP mailbox referrals to indicate to the client that the mailbox is
      //    hosted on a remote server.
      return FromReferral(status, out referrals);
    }

    public static bool FromAppendNoResponse(ImapStatusResponse status, out bool? tryCreate, out ImapUri[] referrals)
    {
      tryCreate = null;
      referrals = null;

      if (status.Status != ImapProtocolStatus.No)
        return false;

      /*
      If the destination mailbox does not exist, a server MUST return an
      error, and MUST NOT automatically create the mailbox.  Unless it
      is certain that the destination mailbox can not be created, the
      server MUST send the response code "[TRYCREATE]" as the prefix of
      the text of the tagged NO response.  This gives a hint to the
      client that it can attempt a CREATE command and retry the APPEND
      if the CREATE is successful.
      */
      if (FromTryCreate(status)) {
        tryCreate = true;
        return true;
      }

      // RFC 2193 IMAP4 Mailbox Referrals
      // http://tools.ietf.org/html/rfc2193
      // 4.1. SELECT, EXAMINE, DELETE, SUBSCRIBE, UNSUBSCRIBE, STATUS and APPEND
      //      Referrals
      //    An IMAP4 server MAY respond to the SELECT, EXAMINE, DELETE,
      //    SUBSCRIBE, UNSUBSCRIBE, STATUS or APPEND command with one or more
      //    IMAP mailbox referrals to indicate to the client that the mailbox is
      //    hosted on a remote server.
      return FromReferral(status, out referrals);
    }

    public static bool FromCreateNoResponse(ImapStatusResponse status, out ImapUri[] referrals)
    {
      referrals = null;

      if (status.Status != ImapProtocolStatus.No)
        return false;

      // RFC 2193 IMAP4 Mailbox Referrals
      // http://tools.ietf.org/html/rfc2193
      // 4.2. CREATE Referrals
      //    An IMAP4 server MAY respond to the CREATE command with one or more
      //    IMAP mailbox referrals, if it wishes to direct the client to issue
      //    the CREATE against another server.  The server can employ any means,
      //    such as examining the hierarchy of the specified mailbox name, in
      //    determining which server the mailbox should be created on.
      return FromReferral(status, out referrals);
    }

    public static bool FromRenameNoResponse(ImapStatusResponse status, out ImapUri[] referrals)
    {
      referrals = null;

      if (status.Status != ImapProtocolStatus.No)
        return false;

      // RFC 2193 IMAP4 Mailbox Referrals
      // http://tools.ietf.org/html/rfc2193
      // 4.3. RENAME Referrals
      //    An IMAP4 server MAY respond to the RENAME command with one or more
      //    pairs of IMAP mailbox referrals.  In each pair of IMAP mailbox
      //    referrals, the first one is an URL to the existing mailbox name and
      //    the second is an URL to the requested new mailbox name.
      return FromReferral(status, out referrals);
    }

    public static bool FromCopyNoResponse(ImapStatusResponse status, out bool? tryCreate, out ImapUri[] referrals)
    {
      tryCreate = null;
      referrals = null;

      if (status.Status != ImapProtocolStatus.No)
        return false;

      /*
      If the destination mailbox does not exist, a server SHOULD return
      an error.  It SHOULD NOT automatically create the mailbox.  Unless
      it is certain that the destination mailbox can not be created, the
      server MUST send the response code "[TRYCREATE]" as the prefix of
      the text of the tagged NO response.  This gives a hint to the
      client that it can attempt a CREATE command and retry the COPY if
      the CREATE is successful.
      */
      if (FromTryCreate(status)) {
        tryCreate = true;
        return true;
      }

      // RFC 2193 IMAP4 Mailbox Referrals
      // http://tools.ietf.org/html/rfc2193
      // 4.4. COPY Referrals
      //    An IMAP4 server MAY respond to the COPY command with one or more IMAP
      //    mailbox referrals.  This indicates that the destination mailbox is on
      //    a remote server.  To achieve the same behavior of a server COPY, the
      //    client MAY issue the constituent FETCH and APPEND commands against
      //    both servers.
      return FromReferral(status, out referrals);
    }

    /*
     * utility methods
     */

    private static bool IsRequiredStatusResponse(ImapStatusResponse status, string code)
    {
      if (status.ResponseCode == null)
        return false;

      if (!status.ResponseCode.Code.Equals(code, StringComparison.InvariantCultureIgnoreCase))
        return false;

      return true;
    }
  }
}