using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Net.Security;
using System.IO;
using System.Security.Authentication;
using System.Security.Cryptography.X509Certificates;
using System.Threading;

using Smdn.Protocols.Imap4.Transaction;

namespace Smdn.Protocols.Imap4.Session {
  // Layers:
  //   Client.ImapClient
  //   Client.ImapClientAgent
  // * Session.ImapSessionAgent
  //   Session.ImapSession
  //   Transaction.ImapTransactionAgent
  //   Transaction.ImapTransaction/Transaction.ImapRequest/Transaction.ImapResponse
  //   Protocol.ImapConnection/Protocol.ImapProtocolCommand/Protocol.ImapProtocolResponse
  //   (IMAP4)

  public class ImapSessionAgent : IDisposable {
    public X509Certificate2Collection CertCollection {
      get { return certCollection; }
    }

    public RemoteCertificateValidationCallback ServerCertificateValidationCallback {
      get { return serverCertificateValidationCallback; }
      set { serverCertificateValidationCallback = value; }
    }

    public ImapCreateAuthenticatedStreamCallback CreateAuthenticatedStreamCallback {
      get { return createAuthenticatedStreamCallback; }
      set { createAuthenticatedStreamCallback = value; }
    }

    public ImapCreateAuthenticatedStreamCallback UpgradeAuthenticatedStreamCallback {
      get { return upgradeAuthenticatedStreamCallback; }
      set { upgradeAuthenticatedStreamCallback = value; }
    }
    /*
     * construction/destruction
     */

    public ImapSessionAgent()
    {
    }

    public void Initialize()
    {
      if (agent != null)
        throw new InvalidOperationException("already initialized");

      Trace.Info("ImapSessionAgent initializing");

      agent = new ImapTransactionAgent();
      agent.Initialize();

      Trace.Info("ImapSessionAgent initialized");
    }

    ~ImapSessionAgent()
    {
      Dispose();
      GC.SuppressFinalize(this);
    }

    public void Dispose()
    {
      if (disposed)
        return;

      Trace.Info("ImapSessionAgent disposing");

      if (agent != null) {
        List<ImapSession> existSessions;

        lock (sessions) {
          existSessions = new List<ImapSession>(sessions.Values);
        }

        foreach (var session in existSessions) {
          if (session.State == ImapSessionState.Selected)
            Close(session);
          if (session.State == ImapSessionState.Authenticated)
            Logout(session);
          if (session.State != ImapSessionState.NotConnected)
            Disconnect(session);
        }

        lock (sessions) {
          sessions.Clear();
        }

        agent.Dispose();
        agent = null;
      }

      Trace.Info("ImapSessionAgent disposed");

      disposed = true;
    }

    /*
     * methods for requesting by IMAP URL
     */

    /// <summary>sends request by IMAP URL</summary>
    public bool Request(ImapSession session, ImapUri requestUri, out ImapResponse response)
    {
      response = null;

      RejectNonConnectedState(session);

      if (session != GetExistSession(requestUri))
        throw new InvalidOperationException("the host specified with IMAP URL is different from the session");

      TraceVerbose(session, "requesting by IMAP URL '{0}'", requestUri);

      switch (requestUri.Form) {
        case ImapUriForm.Server:
        case ImapUriForm.ListMailboxes:
          return ListMailboxes(session, requestUri, out response);

        case ImapUriForm.ListMessages:
          return ListMessages(session, requestUri, out response);

        case ImapUriForm.FetchMessage:
          return FetchMessage(session, requestUri, out response);

        case ImapUriForm.SearchMessages:
          return SearchMessages(session, requestUri, out response);

        default:
          throw new UriFormatException();
      }
    }

    private bool ListMailboxes(ImapSession session, ImapUri requestUri, out ImapResponse response)
    {
      RejectNonAuthenticatedState(session);

      response = null;

      ImapMailbox[] mailboxes;
      ImapTransactionMethod listMethod;

      switch (requestUri.ListType) {
        case ImapListType.List:
          listMethod = ImapTransactionMethod.List;
          break;
        case ImapListType.Lsub:
          listMethod = ImapTransactionMethod.Lsub;
          break;
        default:
          listMethod = ImapTransactionMethod.Lsub;
          break;
      }

      if (!ListLsubCore(listMethod, session, requestUri.Mailbox, out mailboxes))
        return false;

      response = new ImapListResponse(ImapTransactionStatus.Ok, mailboxes, "List succeeded");

      return true;
    }

    private bool ListMessages(ImapSession session, ImapUri requestUri, out ImapResponse response)
    {
      /*
      RFC 2683 IMAP4 Implementation Recommendations
      http://tools.ietf.org/html/rfc2683

      3.2.1.2. Fetching the List of Messages
         When a client selects a mailbox, it is given a count, in the untagged
         EXISTS response, of the messages in the mailbox.  This number can be
         very large.  In such a case it might be unwise to use

             C: 004 FETCH 1:* ALL

         to populate the user's view of the mailbox.  One good method to avoid
         problems with this is to batch the requests, thus:

             C: 004 FETCH 1:50 ALL
             S: * 1 FETCH ...etc...
             S: 004 OK done
             C: 005 FETCH 51:100 ALL
             S: * 51 FETCH ...etc...
             S: 005 OK done
             C: 006 FETCH 101:150 ALL
             ...etc...
      */

      response = null;

      var selectedMailbox = session.SelectedMailbox;

      if (!Status(session, selectedMailbox, ImapStatusDataItem.All))
        return false;

      var messageCount = selectedMailbox.ExistsMessage;
      var fetchedMessages = new List<ImapMessage>();

      for (var fetching = 1; fetching <= messageCount; ) {
        const int fetchUpTo = 20;
        ImapSequenceSet fetchingSet;

        if (fetching + fetchUpTo <= messageCount)
          fetchingSet = ImapSequenceSet.Range(fetching, fetching + fetchUpTo - 1);
        else
          fetchingSet = ImapSequenceSet.From(fetching);

        fetching += fetchUpTo;

        ImapMessage[] messages;

        if (!Fetch(session, fetchingSet, ImapFetchDataItem.Full + ImapFetchDataItem.Uid, out messages))
          return false;

        fetchedMessages.AddRange(messages);
      }

      response = new ImapMessageResponse(ImapTransactionStatus.Ok, fetchedMessages.ToArray(), "Fetch succeeded"); 

      return true;
    }

    private bool FetchMessage(ImapSession session, ImapUri requestUri, out ImapResponse response)
    {
      /*
      RFC 2683 IMAP4 Implementation Recommendations
      http://tools.ietf.org/html/rfc2683

      3.2.1.3. Fetching a Large Body Part

         The issue here is similar to the one for a list of messages.  In the
         BODYSTRUCTURE response the client knows the size, in bytes, of the
         body part it plans to fetch.  Suppose this is a 70 MB video clip. The
         client can use partial fetches to retrieve the body part in pieces,
         avoiding the problem of an uninterruptible 70 MB literal coming back
         from the server:

             C: 022 FETCH 3 BODY[1]<0.20000>
             S: * 3 FETCH (FLAGS(\Seen) BODY[1]<0> {20000}
             S: ...data...)
             S: 022 OK done
             C: 023 FETCH 3 BODY[1]<20001.20000>
             S: * 3 FETCH (BODY[1]<20001> {20000}
             S: ...data...)
             S: 023 OK done
             C: 024 FETCH 3 BODY[1]<40001.20000>
             ...etc...
      */

      response = null;

      // fetch size
      var fetchingSet = ImapSequenceSet.Set(requestUri.MessageUid);
      ImapMessage[] messages;

      if (!UidFetch(session, fetchingSet, ImapFetchDataItem.Rfc822Size, out messages))
        return false;

      if (messages.Length != 1)
        return false;

      var totalOctets = messages[0].BodySize;

      // fetch expect to body
      if (!UidFetch(session, fetchingSet, ImapFetchDataItem.Full + ImapFetchDataItem.Uid, out messages))
        return false;

      if (messages.Length != 1)
        return false;

      ImapMessage fetched = messages[0];

      // fetch body each 10240 octets
      const int octetsUpTo = 10240;
      fetched.Body = string.Empty;

      for (var octet = 0; octet < totalOctets; octet += octetsUpTo) {
        if (!UidFetch(session, fetchingSet, ImapFetchDataItem.BodyPeek(requestUri.Section, octet, octetsUpTo), out messages))
          return false;

        if (messages.Length != 1)
          return false;

        fetched.Body += messages[0].Body;
      }

      response = new ImapMessageResponse(ImapTransactionStatus.Ok, new[] {fetched}, "Fetch succeeded"); 

      return true;
    }

    private bool SearchMessages(ImapSession session, ImapUri requestUri, out ImapResponse response)
    {
      response = null;

      int[] messageSequences;

      if (!Search(session, requestUri.GetSearchCriteria(), out messageSequences))
        return false;

      // RFC 2683 IMAP4 Implementation Recommendations
      // http://tools.ietf.org/html/rfc2683
      // 3.2.1.2. Fetching the List of Messages
      var fetchedMessages = new List<ImapMessage>();
      var fetchMax = messageSequences.Length;

      for (var fetching = 0; fetching < fetchMax; ) {
        const int fetchUpTo = 20;
        ImapSequenceSet fetchingSet;

        if (fetching + fetchUpTo < fetchMax)
          fetchingSet = ImapSequenceSet.Set(messageSequences, fetching, fetchUpTo);
        else
          fetchingSet = ImapSequenceSet.Set(messageSequences, fetching, fetchMax - fetching);

        fetching += fetchUpTo;

        ImapMessage[] messages;

        if (!Fetch(session, fetchingSet, ImapFetchDataItem.Full + ImapFetchDataItem.Uid, out messages))
          return false;

        fetchedMessages.AddRange(messages);
      }

      response = new ImapMessageResponse(ImapTransactionStatus.Ok, fetchedMessages.ToArray(), "Search succeeded");

      return true;
    }

    /*
     * method for connect
     */

    /// <summary>connects to IMAP server</summary>
    public bool Connect(ImapUri uri, out ImapSession session)
    {
      return Connect(uri, false, out session);
    }

    /// <summary>connects to IMAP server</summary>
    public bool Connect(ImapUri uri, bool handlesReferralAsException, out ImapSession session)
    {
      RejectNonInitialized();

      session = GetExistSession(uri);

      if (session != null)
        return true;

      session = ImapSession.Start(uri, CreateAuthenticatedStream);

      if (session == null)
        return false;

      session.HandlesReferralAsException = handlesReferralAsException;

      TraceVerbose(session, "waiting for initial server greeting response");

      ImapResponse response;

      try {
        if (!RequestSync(session, ImapTransactionMethod.Greeting, out response)) {
          session = null;
          return false;
        }
      }
      catch (ImapReferralExcetion) {
        session = null;
        throw;
      }

      lock (sessions) {
        sessions.Add(uri.UserAtHost, session);
      }

      if (response.Status == ImapTransactionStatus.PreAuthed)
        TransitStateTo(session, ImapSessionState.Authenticated);
      else
        TransitStateTo(session, ImapSessionState.NonAuthenticated);

      return true;
    }

    /*
     * requesting methods for any state
     */

    /// <summary>sends CAPABILITY command</summary>
    /// <remarks>valid in any state</remarks>
    public bool Capability(ImapSession session)
    {
      ImapCapability capa;

      return Capability(session, out capa);
    }

    /// <summary>sends CAPABILITY command</summary>
    /// <remarks>valid in any state</remarks>
    public bool Capability(ImapSession session, out ImapCapability capability)
    {
      RejectNonConnectedState(session);

      capability = null;

      ImapResponse response;

      if (RequestSync(session, ImapTransactionMethod.Capability, out response)) {
        var resp = (response as ImapCapabilityResponse);

        capability = resp.Capability;

        return true;
      }

      return false;
    }

    /// <summary>sends NOOP command</summary>
    /// <remarks>valid in any state</remarks>
    public bool NoOp(ImapSession session)
    {
      ImapSizeStatus status;

      return NoOp(session, out status);
    }

    /// <summary>sends NOOP command</summary>
    /// <remarks>valid in any state</remarks>
    public bool NoOp(ImapSession session, out ImapSizeStatus status)
    {
      RejectNonConnectedState(session);

      return NoOpCheckCore(ImapTransactionMethod.NoOp, session, out status);
    }

    private bool NoOpCheckCore(ImapTransactionMethod method, ImapSession session, out ImapSizeStatus status)
    {
      status = null;
        
      ImapResponse response;

      if (RequestSync(session, method, out response)) {
        status = (response as ImapSizeStatusUpdatedResponse).UpdatedStatus;
        return true;
      }

      return false;
    }

    /// <summary>sends LOGOUT command</summary>
    /// <remarks>valid in any state</remarks>
    public bool Logout(ImapSession session)
    {
      if (session.State == ImapSessionState.NotConnected) {
        TraceVerbose(session, "already logged out or disconnected");
        return true;
      }

      if (session.State == ImapSessionState.NonAuthenticated) {
        TraceVerbose(session, "not logged in");
        return true;
      }

      ImapResponse response;

      RequestSync(session, ImapTransactionMethod.Logout, out response);

      if (response.Status == ImapTransactionStatus.Ok) {
        TransitStateTo(session, ImapSessionState.NonAuthenticated);
        Disconnect(session);
        return true;
      }
      else if (response.Status == ImapTransactionStatus.Disconnecting) {
        return true;
      }
      else {
        return false;
      }
    }

    /*
     * requesting methods for non authenticated state
     */

    /// <summary>sends STARTTLS command</summary>
    /// <remarks>valid in non authenticated state</remarks>
    public bool StartTls(ImapSession session)
    {
      RejectNonConnectedState(session);

      if (session.State != ImapSessionState.NonAuthenticated) {
        TraceVerbose(session, "already authenticated");
        return true;
      }

      return RequestSync(session, new ImapStartTlsRequest(UpgradeStream, session.Timeout));
    }

    /// <summary>sends AUTHENTICATE command</summary>
    /// <remarks>valid in non authenticated state</remarks>
    public bool Authenticate(ImapSession session, string password)
    {
      RejectNonConnectedState(session);

      if (session.State != ImapSessionState.NonAuthenticated) {
        TraceVerbose(session, "already authenticated");
        return true;
      }

      var request = new ImapRequest(ImapTransactionMethod.Authenticate, session.Timeout);

      request.Arguments["user name"] = session.HostUri.UserName;
      request.Arguments["password"] = password;
      request.Arguments["authentication mechanism name"] = ImapConverter.ToString(session.HostUri.Authentication);

      if (RequestSync(session, request)) {
        TransitStateTo(session, ImapSessionState.Authenticated);
        return true;
      }

      return false;
    }

    /// <summary>sends LOGIN command</summary>
    /// <remarks>valid in non authenticated state</remarks>
    public bool Login(ImapSession session, string password)
    {
      RejectNonConnectedState(session);

      if (session.State != ImapSessionState.NonAuthenticated) {
        TraceVerbose(session, "already logged in");
        return true;
      }

      var request = new ImapRequest(ImapTransactionMethod.Login, session.Timeout);

      request.Arguments["user name"] = session.HostUri.UserName;
      request.Arguments["password"] = password;

      if (RequestSync(session, request)) {
        TransitStateTo(session, ImapSessionState.Authenticated);
        return true;
      }

      return false;
    }

    /*
     * requesting methods for authenticated state
     */

    /// <summary>sends SELECT command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public bool Select(ImapSession session, ImapMailbox mailbox)
    {
      ValidateMailboxRelationship(session, mailbox);

      return SelectExamineCore(ImapTransactionMethod.Select, session, mailbox.Name);
    }

    /// <summary>sends SELECT command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public bool Select(ImapSession session, string mailboxName)
    {
      return SelectExamineCore(ImapTransactionMethod.Select, session, mailboxName);
    }

    /// <summary>sends EXAMINE command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public bool Examine(ImapSession session, ImapMailbox mailbox)
    {
      ValidateMailboxRelationship(session, mailbox);

      return SelectExamineCore(ImapTransactionMethod.Examine, session, mailbox.Name);
    }

    /// <summary>sends EXAMINE command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public bool Examine(ImapSession session, string mailboxName)
    {
      return SelectExamineCore(ImapTransactionMethod.Examine, session, mailboxName);
    }

    private bool SelectExamineCore(ImapTransactionMethod method, ImapSession session, string mailboxName)
    {
      RejectNonAuthenticatedState(session);

      var existMailbox = ImapMailboxManager.GetExistMailbox(session.ServerInfo, mailboxName);

      if (existMailbox != null) {
        ValidateMailboxRelationship(session, existMailbox);

        if (session.SelectedMailbox == existMailbox) {
          TraceVerbose(session, "mailbox has already been selected");
          return true;
        }
        else if (existMailbox.Flags != null &&
            existMailbox.Flags.Has(ImapKnownMailboxFlags.NoSelect)) {
          throw new InvalidOperationException("mailbox is not selectable");
        }
      }

      var request = new ImapRequest(method, session.Timeout);

      request.Arguments["mailbox name"] = mailboxName;

      ImapResponse response;

      if (RequestSync(session, request, out response)) {
        session.SelectedMailbox = (response as ImapMailboxResponse).Mailbox;
        TransitStateTo(session, ImapSessionState.Selected);
        return true;
      }

      return false;
    }

    /// <summary>sends CREATE command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public bool Create(ImapSession session, string mailboxName, out ImapMailbox createdMailbox)
    {
      RejectNonAuthenticatedState(session);

      createdMailbox = null;

      var request = new ImapRequest(ImapTransactionMethod.Create, session.Timeout);

      request.Arguments["mailbox name"] = mailboxName;

      ImapResponse response;

      if (RequestSync(session, request, out response)) {
        createdMailbox = (response as ImapMailboxResponse).Mailbox;
        return true;
      }

      return false;
    }

    /// <summary>sends DELETE command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public bool Delete(ImapSession session, ImapMailbox mailbox)
    {
      ValidateMailboxRelationship(session, mailbox);

      return Delete(session, mailbox.Name);
    }

    /// <summary>sends DELETE command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public bool Delete(ImapSession session, string mailboxName)
    {
      RejectNonAuthenticatedState(session);

      var request = new ImapRequest(ImapTransactionMethod.Delete, session.Timeout);

      request.Arguments["mailbox name"] = mailboxName;

      return RequestSync(session, request);
    }

    /// <summary>sends RENAME command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public bool Rename(ImapSession session, ImapMailbox existingMailbox, string newMailboxName, out ImapMailbox renamedMailbox)
    {
      ValidateMailboxRelationship(session, existingMailbox);

      return Rename(session, existingMailbox.Name, newMailboxName, out renamedMailbox);
    }

    /// <summary>sends RENAME command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public bool Rename(ImapSession session, string existingMailboxName, string newMailboxName, out ImapMailbox renamedMailbox)
    {
      RejectNonAuthenticatedState(session);

      renamedMailbox = null;

      var request = new ImapRequest(ImapTransactionMethod.Rename, session.Timeout);

      request.Arguments["existing mailbox name"] = existingMailboxName;
      request.Arguments["new mailbox name"] = newMailboxName;

      ImapResponse response;

      if (RequestSync(session, request, out response)) {
        renamedMailbox = (response as ImapMailboxResponse).Mailbox;
        return true;
      }

      return false;
    }

    /// <summary>sends SUBSCRIBE command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public bool Subscribe(ImapSession session, ImapMailbox mailbox)
    {
      ValidateMailboxRelationship(session, mailbox);

      return Subscribe(session, mailbox.Name);
    }

    /// <summary>sends SUBSCRIBE command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public bool Subscribe(ImapSession session, string mailboxName)
    {
      RejectNonAuthenticatedState(session);

      var request = new ImapRequest(ImapTransactionMethod.Subscribe, session.Timeout);

      request.Arguments["mailbox name"] = mailboxName;

      return RequestSync(session, request);
    }

    /// <summary>sends UNSUBSCRIBE command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public bool Unsubscribe(ImapSession session, ImapMailbox mailbox)
    {
      ValidateMailboxRelationship(session, mailbox);

      return Unsubscribe(session, mailbox.Name);
    }

    /// <summary>sends UNSUBSCRIBE command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public bool Unsubscribe(ImapSession session, string mailboxName)
    {
      RejectNonAuthenticatedState(session);

      var request = new ImapRequest(ImapTransactionMethod.Unsubscribe, session.Timeout);

      request.Arguments["mailbox name"] = mailboxName;

      return RequestSync(session, request);
    }

    /// <summary>sends LIST command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public bool List(ImapSession session, string mailboxName, out ImapMailbox[] mailboxes)
    {
      return ListLsubCore(ImapTransactionMethod.List, session, mailboxName, out mailboxes);
    }

    /// <summary>sends LSUB command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public bool Lsub(ImapSession session, string mailboxName, out ImapMailbox[] mailboxes)
    {
      return ListLsubCore(ImapTransactionMethod.Lsub, session, mailboxName, out mailboxes);
    }

    private bool ListLsubCore(ImapTransactionMethod listMethod, ImapSession session, string mailboxName, out ImapMailbox[] mailboxes)
    {
      RejectNonAuthenticatedState(session);

      mailboxes = null;

      var request = new ImapRequest(listMethod, session.Timeout);

      request.Arguments["mailbox name"] = mailboxName;

      ImapResponse response;

      if (RequestSync(session, request, out response)) {
        mailboxes = (response as ImapListResponse).Mailboxes;
        return true;
      }

      return false;
    }

    /// <summary>sends STATUS command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public bool Status(ImapSession session, ImapMailbox mailbox, ImapStatusDataItem statusDataItem)
    {
      ValidateMailboxRelationship(session, mailbox);

      ImapMailbox updatedMailbox;

      Status(session, mailbox.Name, statusDataItem, out updatedMailbox);
 
      // 常に同じインスタンスとなるはず
      return (mailbox == updatedMailbox);
    }

    /// <summary>sends STATUS command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public bool Status(ImapSession session, string mailboxName, ImapStatusDataItem statusDataItem, out ImapMailbox mailbox)
    {
      RejectNonAuthenticatedState(session);

      mailbox = null;

      var request = new ImapRequest(ImapTransactionMethod.Status, session.Timeout);

      // mailbox name
      request.Arguments["mailbox name"] = mailboxName;

      // status data item names
      request.Arguments["status data item names"] = statusDataItem.ToStringList();

      ImapResponse response;

      if (RequestSync(session, request, out response)) {
        mailbox = (response as ImapMailboxResponse).Mailbox;
        return true;
      }

      return false;
    }

    /// <summary>sends APPEND command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public bool Append(ImapSession session, ImapMessage message, ImapMailbox mailbox)
    {
      ValidateMailboxRelationship(session, mailbox);

      ImapMailbox created;

      return AppendCore(session, message, mailbox.Name, false, out created);
    }

    /// <summary>sends APPEND command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public bool Append(ImapSession session, ImapMessage message, string mailboxName)
    {
      ImapMailbox created;

      return AppendCore(session, message, mailboxName, false, out created);
    }

    /// <summary>sends APPEND command</summary>
    /// <remarks>
    /// valid in authenticated state,
    /// this method tries to automatically create the mailbox when server sent [TRYCREATE] response code
    /// </remarks>

    /// <remarks>valid in  state, this method tries to automatically create the mailbox when server sent [TRYCREATE] response code</remarks>
    public bool Append(ImapSession session, ImapMessage message, string mailboxName, out ImapMailbox createdMailbox)
    {
      return AppendCore(session, message, mailboxName, true, out createdMailbox);
    }

    private bool AppendCore(ImapSession session, ImapMessage message, string mailboxName, bool tryCreate, out ImapMailbox createdMailbox)
    {
      RejectNonAuthenticatedState(session);

      createdMailbox = null;

      var request = new ImapRequest(ImapTransactionMethod.Append, session.Timeout);

      // mailbox name
      request.Arguments["mailbox name"] = mailboxName;

      // flag parenthesized list
      if (message.Flags != null)
        request.Arguments["flag parenthesized list"] = ImapConverter.ToFlagList(message.Flags);

      // date/time string
      if (!DateTime.MinValue.Equals(message.InternalDate))
        request.Arguments["date/time string"] = ImapConverter.ToDateTimeString(message.InternalDate);

      // message literal
      request.Arguments["message literal"] = new ImapLiteralString(message.Body);

      ImapResponse response;

      if (RequestSync(session, request, out response)) {
        UpdateSelectedMailboxSizeStatus(session, (response as ImapSizeStatusUpdatedResponse).UpdatedStatus);
        return true;
      }

      // try create
      if (response.Status == ImapTransactionStatus.TryCreate && tryCreate)
        return TryCreateAndRequestAgain(session, request, mailboxName, out response, out createdMailbox);

      return false;
    }

    /*
     * requesting methods for selected state
     */

    /// <summary>sends CHECK command</summary>
    /// <remarks>valid in selected state</remarks>
    public bool Check(ImapSession session)
    {
      ImapSizeStatus status;

      return Check(session, out status);
    }

    /// <summary>sends NOOP command</summary>
    /// <remarks>valid in any state</remarks>
    public bool Check(ImapSession session, out ImapSizeStatus status)
    {
      RejectNonSelectedState(session);

      return NoOpCheckCore(ImapTransactionMethod.Check, session, out status);
    }

    /// <summary>sends CLOSE command</summary>
    /// <remarks>valid in selected state</remarks>
    public bool Close(ImapSession session)
    {
      RejectNonAuthenticatedState(session);

      if (session.State != ImapSessionState.Selected) {
        TraceVerbose(session, "mailbox has already closed or has not selected yet");
        return true;
      }

      if (RequestSync(session, ImapTransactionMethod.Close)) {
        session.SelectedMailbox = null;
        TransitStateTo(session, ImapSessionState.Authenticated);
        return true;
      }

      return false;
    }

    /// <summary>sends EXPUNGE command</summary>
    /// <remarks>valid in selected state</remarks>
    public bool Expunge(ImapSession session)
    {
      int[] expunged;

      return Expunge(session, out expunged);
    }

    /// <summary>sends EXPUNGE command</summary>
    /// <remarks>valid in selected state</remarks>
    public bool Expunge(ImapSession session, out int[] expungedMessages)
    {
      RejectNonSelectedState(session);

      expungedMessages = null;

      ImapResponse response;

      if (RequestSync(session, ImapTransactionMethod.Expunge, out response)) {
        expungedMessages = (response as ImapMessageListResponse).Messages;
        return true;
      }

      return false;
    }

    /// <summary>sends SEARCH command</summary>
    /// <remarks>valid in selected state</remarks>
    public bool Search(ImapSession session, ImapSearchCriteria searchingCriteria, out int[] messages)
    {
      return SearchCore(session, false, searchingCriteria, out messages);
    }

    /// <summary>sends UID SEARCH command</summary>
    /// <remarks>valid in selected state</remarks>
    public bool UidSearch(ImapSession session, ImapSearchCriteria searchingCriteria, out int[] messages)
    {
      return SearchCore(session, true, searchingCriteria, out messages);
    }

    private bool SearchCore(ImapSession session, bool uidSearch, ImapSearchCriteria searchingCriteria, out int[] messages)
    {
      RejectNonSelectedState(session);

      messages = null;

      var request = new ImapRequest(ImapTransactionMethod.Search, session.Timeout);

      // dummy parameter to decide sending UID SEARCH or SEARCH
      if (uidSearch)
        request.Arguments["uid"] = "uid";

      // searching criteria
      request.Arguments["searching criteria"] = searchingCriteria.ToStringList();

      ImapResponse response;

      if (RequestSync(session, request, out response)) {
        messages = (response as ImapMessageListResponse).Messages;
        return true;
      }

      return false;
    }

    /// <summary>sends FETCH command</summary>
    /// <remarks>valid in selected state</remarks>
    public bool Fetch(ImapSession session, ImapSequenceSet sequenceSet, ImapFetchDataItem messageDataItems, out ImapMessage[] messages)
    {
      return FetchCore(session, sequenceSet, false, messageDataItems, out messages);
    }

    /// <summary>sends UID FETCH command</summary>
    /// <remarks>valid in selected state</remarks>
    public bool UidFetch(ImapSession session, ImapSequenceSet uidSet, ImapFetchDataItem messageDataItems, out ImapMessage[] messages)
    {
      return FetchCore(session, uidSet, true, messageDataItems, out messages);
    }

    private bool FetchCore(ImapSession session, ImapSequenceSet fetchingSequenceSet, bool uidSet, ImapFetchDataItem messageDataItems, out ImapMessage[] messages)
    {
      RejectNonSelectedState(session);

      messages = null;

      var request = new ImapRequest(ImapTransactionMethod.Fetch, session.Timeout);

      // sequence set
      request.Arguments[uidSet ? "uid set" : "sequence set"] = fetchingSequenceSet.ToString();

      // message data item names
      request.Arguments["message data item names"] = messageDataItems.ToStringList();

      ImapResponse response;

      if (RequestSync(session, request, out response)) {
        messages = (response as ImapMessageResponse).Messages;

        foreach (var message in messages) {
          message.Mailbox = session.SelectedMailbox;
        }

        return true;
      }

      return false;
    }

    /// <summary>sends STORE command</summary>
    /// <remarks>valid in selected state</remarks>
    public bool Store(ImapSession session, ImapSequenceSet sequenceSet, ImapStoreDataItem messageDataItem)
    {
      ImapMessage[] messages;

      return StoreCore(session, sequenceSet, false, messageDataItem, out messages);
    }

    /// <summary>sends STORE command</summary>
    /// <remarks>valid in selected state</remarks>
    public bool Store(ImapSession session, ImapSequenceSet sequenceSet, ImapStoreDataItem messageDataItem, out ImapMessage[] messages)
    {
      return StoreCore(session, sequenceSet, false, messageDataItem, out messages);
    }

    /// <summary>sends UID STORE command</summary>
    /// <remarks>valid in selected state</remarks>
    public bool UidStore(ImapSession session, ImapSequenceSet uidSet, ImapStoreDataItem messageDataItem)
    {
      ImapMessage[] messages;

      return StoreCore(session, uidSet, true, messageDataItem, out messages);
    }

    /// <summary>sends UID STORE command</summary>
    /// <remarks>valid in selected state</remarks>
    public bool UidStore(ImapSession session, ImapSequenceSet uidSet, ImapStoreDataItem messageDataItem, out ImapMessage[] messages)
    {
      return StoreCore(session, uidSet, true, messageDataItem, out messages);
    }

    private bool StoreCore(ImapSession session, ImapSequenceSet storingSequenceSet, bool uidSet, ImapStoreDataItem messageDataItem, out ImapMessage[] messages)
    {
      RejectNonSelectedState(session);

      messages = null;

      var request = new ImapRequest(ImapTransactionMethod.Store, session.Timeout);

      // sequence set
      request.Arguments[uidSet ? "uid set" : "sequence set"] = storingSequenceSet.ToString();

      // message data item names/value for message data item
      request.Arguments["message data item name"] = messageDataItem.ToStringList();

      ImapResponse response;

      if (RequestSync(session, request, out response)) {
        messages = (response as ImapMessageResponse).Messages;

        foreach (var message in messages) {
          message.Mailbox = session.SelectedMailbox;
        }

        return true;
      }

      return false;
    }

    /// <summary>sends COPY command</summary>
    /// <remarks>valid in selected state</remarks>
    public bool Copy(ImapSession session, ImapSequenceSet sequenceSet, ImapMailbox mailbox)
    {
      ValidateMailboxRelationship(session, mailbox);

      ImapMailbox created;

      return CopyCore(session, sequenceSet, false, mailbox.Name, false, out created);
    }

    /// <summary>sends COPY command</summary>
    /// <remarks>valid in selected state</remarks>
    public bool Copy(ImapSession session, ImapSequenceSet sequenceSet, string mailboxName)
    {
      ImapMailbox created;

      return CopyCore(session, sequenceSet, false, mailboxName, false, out created);
    }

    /// <summary>sends COPY command</summary>
    /// <remarks>
    /// valid in selected state,
    /// this method tries to automatically create the mailbox when server sent [TRYCREATE] response code
    /// </remarks>
    public bool Copy(ImapSession session, ImapSequenceSet sequenceSet, string mailboxName, out ImapMailbox createdMailbox)
    {
      return CopyCore(session, sequenceSet, false, mailboxName, true, out createdMailbox);
    }

    /// <summary>sends UID COPY command</summary>
    /// <remarks>valid in selected state</remarks>
    public bool UidCopy(ImapSession session, ImapSequenceSet uidSet, ImapMailbox mailbox)
    {
      ValidateMailboxRelationship(session, mailbox);

      ImapMailbox created;

      return CopyCore(session, uidSet, true, mailbox.Name, false, out created);
    }

    /// <summary>sends UID COPY command</summary>
    /// <remarks>valid in selected state</remarks>
    public bool UidCopy(ImapSession session, ImapSequenceSet uidSet, string mailboxName)
    {
      ImapMailbox created;

      return CopyCore(session, uidSet, true, mailboxName, false, out created);
    }

    /// <summary>sends UID COPY command</summary>
    /// <remarks>
    /// valid in selected state,
    /// this method tries to automatically create the mailbox when server sent [TRYCREATE] response code
    /// </remarks>
    public bool UidCopy(ImapSession session, ImapSequenceSet uidSet, string mailboxName, out ImapMailbox createdMailbox)
    {
      return CopyCore(session, uidSet, true, mailboxName, true, out createdMailbox);
    }

    private bool CopyCore(ImapSession session, ImapSequenceSet copyingSequenceSet, bool uidSet, string mailboxName, bool tryCreate, out ImapMailbox createdMailbox)
    {
      RejectNonSelectedState(session);

      createdMailbox = null;

      var request = new ImapRequest(ImapTransactionMethod.Copy, session.Timeout);

      // sequence set
      request.Arguments[uidSet ? "uid set" : "sequence set"] = copyingSequenceSet.ToString();

      // mailbox name
      request.Arguments["mailbox name"] = mailboxName;

      ImapResponse response;

      if (RequestSync(session, request, out response))
        return true;

      // try create
      if (response.Status == ImapTransactionStatus.TryCreate && tryCreate)
        return TryCreateAndRequestAgain(session, request, mailboxName, out response, out createdMailbox);

      return false;
    }

    /// <summary>sends UNSELECT command</summary>
    /// <remarks>valid in selected state</remarks>
    public bool Unselect(ImapSession session)
    {
      RejectNonSelectedState(session);

      if (RequestSync(session, ImapTransactionMethod.Unselect)) {
        session.SelectedMailbox = null;
        TransitStateTo(session, ImapSessionState.Authenticated);
        return true;
      }

      return false;
    }

    /*
     * requesting methods for authenticated/selected state
     */

    /// <summary>sends IDLE command</summary>
    /// <remarks>valid in authenticated/selected state</remarks>
    public bool BeginIdle(ImapSession session, ImapSizeStatusUpdatedCallback callback)
    {
      RejectNonAuthenticatedState(session);

      if (session.Connection.IsIdling)
        throw new InvalidOperationException("already idling");

      session.Connection.IdlingStatusUpdatedCallback = callback;

      var request = new ImapRequest(ImapTransactionMethod.Idle, session.Timeout);

      var succeeded = true;
      var transaction = RequestAsync(session, request);

      for ( ; ; Thread.Sleep(50)) {
        if (session.Connection.IsIdling)
          break; // succeeded
        if (transaction.Response != null &&
            transaction.Response.Status != ImapTransactionStatus.Ok) {
          succeeded = false; // failed
          break;
        }
        if (request.Timeout < DateTime.Now.Subtract(request.RequestedTime).TotalMilliseconds) {
          succeeded = false; // time out
          break;
        }
      }

      if (succeeded)
        TraceVerbose(session, "session is idling now");
      else
        session.Connection.IdlingStatusUpdatedCallback = null;

      return succeeded;
    }

    /// <summary>sends DONE for IDLE command continuation request responses</summary>
    /// <remarks>valid in idling</remarks>
    public bool EndIdle(ImapSession session)
    {
      ImapSizeStatus status;

      return EndIdle(session, out status);
    }

    /// <summary>sends DONE for IDLE command continuation request responses</summary>
    /// <remarks>valid in idling</remarks>
    public bool EndIdle(ImapSession session, out ImapSizeStatus status)
    {
      RejectNonAuthenticatedState(session);

      status = null;

      if (!session.Connection.IsIdling) {
        TraceVerbose(session, "not idled");
        return true;
      }

      var request = new ImapRequest(ImapTransactionMethod.IdleDone, session.Timeout);

      if (!RequestSync(session, request))
        return false; // DONE failed

      var succeeded = true;

      for ( ; ; Thread.Sleep(50)) {
        if (!session.Connection.IsIdling)
          break; // succeeded
        if (request.Timeout < DateTime.Now.Subtract(request.RequestedTime).TotalMilliseconds) {
          succeeded = false;  // time out
          break;
        }
        agent.Signal();
      }

      status = session.Connection.UpdatedStatusWhileIdling;

      UpdateSelectedMailboxSizeStatus(session, status);

      if (succeeded) {
        TraceVerbose(session, "idle completed");
        session.Connection.IdlingStatusUpdatedCallback = null;
      }

      return succeeded;
    }

    /// <summary>sends NAMESPACE command</summary>
    /// <remarks>valid in authenticated/selected state</remarks>
    public bool Namespace(ImapSession session, out ImapNamespace @namespace)
    {
      RejectNonAuthenticatedState(session);

      @namespace = null;
      
      ImapResponse response;

      if (RequestSync(session, ImapTransactionMethod.Namespace, out response)) {
        @namespace = (response as ImapNamespaceResponse).Namespace;

        return true;
      }

      return false;
    }

    /*
     * other requesting methods
     */

    /// <summary>sends XYZZY command</summary>
    /// <remarks>Gimap XYZZY capability extension</remarks>
    public bool Xyzzy(ImapSession session)
    {
      RejectNonConnectedState(session);

      return RequestSync(session, ImapTransactionMethod.Xyzzy);
    }

    /*
     * internal requesting methods
     */
    private ImapTransaction RequestAsync(ImapSession session, ImapRequest request)
    {
      RejectNonInitialized();
      RejectDisposed();

      if (request.Method != ImapTransactionMethod.IdleDone)
        RejectIdling(session);

      // TODO: レスポンスがDisconnectingになった場合
      return agent.Request(session.Connection, request);
    }

    private bool RequestSync(ImapSession session, ImapTransactionMethod method)
    {
      ImapResponse response;

      return RequestSync(session, method, out response);
    }

    private bool RequestSync(ImapSession session, ImapTransactionMethod method, out ImapResponse response)
    {
      return RequestSync(session, new ImapRequest(method, session.Timeout), out response);
    }

    private bool RequestSync(ImapSession session, ImapRequest request)
    {
      ImapResponse response;

      return RequestSync(session, request, out response);
    }

    private bool RequestSync(ImapSession session, ImapRequest request, out ImapResponse response)
    {
      RejectNonInitialized();
      RejectDisposed();

      response = null;

      if (request.Method != ImapTransactionMethod.IdleDone)
        RejectIdling(session);

      for (var retry = 0; retry < 3; retry++) {
        var transaction = agent.Request(session.Connection, request);

        while (transaction.Response == null) {
          Thread.Sleep(20);
        }

        response = transaction.Response;

        if (response.Status == ImapTransactionStatus.TransactionProgress) {
          // retry
          Thread.Sleep(250);
          continue;
        }
        else {
          break;
        }
      }

      if (response.Status == ImapTransactionStatus.Disconnecting) {
        TransitStateTo(session, ImapSessionState.NonAuthenticated);
        Disconnect(session);
      }

      if (response is ImapSizeStatusUpdatedResponse) {
        // update selected mailbox size and status
        UpdateSelectedMailboxSizeStatus(session, (response as ImapSizeStatusUpdatedResponse).UpdatedStatus);
      }
      else if (response is ImapMailboxReferralResponse) {
        // MAILBOX-REFERRAL
        if (session.HandlesReferralAsException) {
          throw new ImapMailboxReferralExcetion(response.Description, (response as ImapMailboxReferralResponse).Referrals);
        }
        else {
          Trace.Info("mailbox referral: '{0}'", response.Description);
          Trace.Info("  try mailboxes below:");
          foreach (var referral in (response as ImapMailboxReferralResponse).Referrals) {
            Trace.Info("    {0}", referral);
          }
        }
      }
      else if (response is ImapLoginReferralResponse) {
        // LOGIN-REFERRAL
        if (session.HandlesReferralAsException) {
          throw new ImapLoginReferralExcetion(response.Description, (response as ImapLoginReferralResponse).Referral);
        }
        else {
          Trace.Info("login referral: '{0}'", response.Description);
          Trace.Info("  try to connect to {0}", (response as ImapLoginReferralResponse).Referral);
        }
      }

      // true for 2xx, false for 3xx/4xx/5xx
      return ((int)response.Status < 300);
    }

    /*
     * internal methods
     */

    internal ImapSession GetExistSession(ImapUri uri)
    {
      lock (sessions) {
        if (sessions.ContainsKey(uri.UserAtHost))
          return sessions[uri.UserAtHost];
      }

      return null;
    }

    /*
     * methods for checking status
     */

    private static void RejectNonConnectedState(ImapSession session)
    {
      if (session == null)
        throw new NullReferenceException("session is null");

      if (session.State == ImapSessionState.NotConnected)
        throw new InvalidOperationException("session has been disconnected");
    }

    private static void RejectNonAuthenticatedState(ImapSession session)
    {
      RejectNonConnectedState(session);

      if (session.State == ImapSessionState.NonAuthenticated)
        throw new InvalidOperationException("session is not in authenticated state");
    }

    private static void RejectNonSelectedState(ImapSession session)
    {
      RejectNonConnectedState(session);

      if (session.State != ImapSessionState.Selected)
        throw new InvalidOperationException("session is not in selected state");
    }

    private static void RejectIdling(ImapSession session)
    {
      if (session.Connection.IsIdling)
        throw new InvalidOperationException("connection is idling now");
    }

    private static void ValidateMailboxRelationship(ImapSession session, ImapMailbox mailbox)
    {
      if (mailbox.ServerInfo != session.Connection.ServerInfo)
        throw new InvalidOperationException("invalid mailbox");
    }

    /*
     * private methods
     */

    private bool TryCreateAndRequestAgain(ImapSession session, ImapRequest originalRequest, string mailboxName, out ImapResponse response, out ImapMailbox createdMailbox)
    {
      response = null;
      createdMailbox = null;

      Trace.Info("trying to create mailbox '{0}'", mailboxName);

      if (!Create(session, mailboxName, out createdMailbox))
        return false;

      Trace.Info("created mailbox '{0}'", createdMailbox.Name);

      return RequestSync(session, originalRequest, out response);
    }

    private void Disconnect(ImapSession session)
    {
      lock (sessions) {
        if (sessions.ContainsKey(session.HostUri.UserAtHost)) {
          sessions.Remove(session.HostUri.UserAtHost);
        }
      }

      session.Close();

      TraceVerbose(session, "now in non-connected state");
    }

    private void TransitStateTo(ImapSession session, ImapSessionState state)
    {
      switch (state) {
        case ImapSessionState.NonAuthenticated:
          TraceVerbose(session, "now in non-authenticated state");
          break;

        case ImapSessionState.Authenticated:
          TraceVerbose(session, "now in authenticated state");
          break;

        case ImapSessionState.Selected:
          TraceVerbose(session, "now in selected state (selected '{0}')", session.SelectedMailbox.Name);
          break;
      }

      session.State = state;
    }

    private void UpdateSelectedMailboxSizeStatus(ImapSession session, ImapSizeStatus status)
    {
      if (session.SelectedMailbox != null) {
        if (status.Exists != null)
          session.SelectedMailbox.ExistsMessage = (int)status.Exists;
        if (status.Recent != null)
          session.SelectedMailbox.RecentMessage = (int)status.Recent;

        Trace.Dump(session.SelectedMailbox);
      }
    }

    private Stream UpgradeStream(ImapUri uri, Stream baseStream)
    {
      if (upgradeAuthenticatedStreamCallback == null) {
        SslStream sslStream = null;

        if (serverCertificateValidationCallback == null)
          sslStream = new SslStream(baseStream, false);
        else
          sslStream = new SslStream(baseStream, false, serverCertificateValidationCallback);

        sslStream.AuthenticateAsClient(uri.Host, certCollection, SslProtocols.Default, true);

        Trace.Verbose("[ImapSession] Created SSL stream for {0}", uri);
        Trace.Verbose("[ImapSession]   IsAuthenticated: {0}", sslStream.IsAuthenticated);
        Trace.Verbose("[ImapSession]   IsMutuallyAuthenticated: {0}", sslStream.IsMutuallyAuthenticated);
        Trace.Verbose("[ImapSession]   IsEncrypted: {0}", sslStream.IsEncrypted);
        Trace.Verbose("[ImapSession]   IsSigned: {0}", sslStream.IsSigned);

        return sslStream;
      }
      else {
        return upgradeAuthenticatedStreamCallback(uri, baseStream);
      }
    }

    private Stream CreateAuthenticatedStream(ImapUri uri, Stream baseStream)
    {
      if (createAuthenticatedStreamCallback == null) {
        const int imapsPort = 993;

        if (uri.Port == imapsPort)
          return UpgradeStream(uri, baseStream);
        else
          return baseStream;
      }
      else {
        return createAuthenticatedStreamCallback(uri, baseStream);
      }
    }

    [Conditional("TRACE")]
    private void TraceVerbose(ImapSession session, string format, params object[] arguments)
    {
      Trace.Verbose("CID:{0} （≧▽≦）[ImapSession] {1}", session.Id, string.Format(format, arguments));
    }

    private void RejectNonInitialized()
    {
      if (agent == null)
        throw new InvalidOperationException("not initialized");
    }

    private void RejectDisposed()
    {
      if (disposed)
        throw new ObjectDisposedException(GetType().Name);
    }

    private bool disposed = false;
    private ImapTransactionAgent agent = null;
    private readonly Dictionary<string, ImapSession> sessions = new Dictionary<string, ImapSession>();
    private readonly X509Certificate2Collection certCollection = new X509Certificate2Collection();
    private RemoteCertificateValidationCallback serverCertificateValidationCallback = null;
    private ImapCreateAuthenticatedStreamCallback createAuthenticatedStreamCallback = null;
    private ImapCreateAuthenticatedStreamCallback upgradeAuthenticatedStreamCallback = null;
  }
}