using System;

using Smdn.Protocols.Imap4.Protocol;

namespace Smdn.Protocols.Imap4.Transaction.Transactions {
  internal class LoginTransaction : TransactionBase {

    // ProcessLoginDisabled
    //   x Conclude
    // ProcessArgumentNotSetted
    //   x Conclude
    // ProcessLogin
    //   ->WaitForLogin
    // WaitForLogin
    //   x Conclude

    public LoginTransaction(ImapConnection connection, ImapRequest request)
      : base(connection, request)
    {
    }

    protected override ProcessTransactionDelegate Reset()
    {
      // RFC 2595 Using TLS with IMAP, POP3 and ACAP
      // http://tools.ietf.org/html/rfc2595
      // 3.2. IMAP LOGINDISABLED capability
      //    An IMAP client which complies with this specification MUST NOT issue
      //    the LOGIN command if this capability is present.
      if (Connection.ServerInfo.Capabilities.Has(ImapKnownCapabilities.LoginDisabled))
        return ProcessLoginDisabled;
      if (!Request.Arguments.ContainsKey("user name"))
        return ProcessArgumentNotSetted;
      if (!Request.Arguments.ContainsKey("password"))
        return ProcessArgumentNotSetted;

      return ProcessLogin;
    }

    private void ProcessLoginDisabled()
    {
      Conclude(new ImapResponse(ImapTransactionStatus.BadRequest, "LOGIN is disabled"));
    }

    private void ProcessArgumentNotSetted()
    {
      Conclude(new ImapResponse(ImapTransactionStatus.BadRequest, "arguments 'user name' and 'password' must be setted"));
    }

    private void ProcessLogin()
    {
      Send("login",
           WaitForLogin,
           Request.Arguments["user name"],
           Request.Arguments["password"]);
    }

    private void WaitForLogin()
    {
      ImapProtocolResponse response;

      if (!Receive(out response))
        return;

      if (response is ImapStatusResponse) {
        var status = response as ImapStatusResponse;

        if (status.Tag != LastTag)
          return;

        ImapUri referral = null;

        ImapStatusResponseConverter.FromLoginResponse(status, out referral);

        if (status.Status == ImapProtocolStatus.Ok) {
          if (referral == null)
            Conclude(new ImapResponse(ImapTransactionStatus.Ok, "logged in"));
          else
            Conclude(new ImapMailboxReferralResponse(ImapTransactionStatus.PersonalMailboxReferral, new[] {referral}, status.Text));
        }
        else if (status.Status == ImapProtocolStatus.No && referral != null) {
          Conclude(new ImapLoginReferralResponse(ImapTransactionStatus.HomeServerReferral, referral, status.Text));
        }
        else {
          ConcludeError(status);
        }
      }
    }

  }
}
