using System;
using System.Collections.Generic;
using System.Text;

using Smdn.Protocols.Imap4.Protocol;

namespace Smdn.Protocols.Imap4.Transaction.Transactions {
  internal class SearchTransaction : TransactionBase {

    // ProcessArgumentNotSetted
    //   x Conclude
    // ProcessSearch
    //   ->WaitForSearch
    // WaitForSearch
    //   ->ProcessSearch(NO [BADCHARSET])
    //   x Conclude

    public SearchTransaction(ImapConnection connection, ImapRequest request)
      : base(connection, request)
    {
    }

    protected override ProcessTransactionDelegate Reset()
    {
      if (!Request.Arguments.ContainsKey("searching criteria"))
        return ProcessArgumentNotSetted;
      else
        return ProcessSearch;
    }


    private void ProcessArgumentNotSetted()
    {
      Conclude(new ImapResponse(ImapTransactionStatus.BadRequest, "arguments 'searching criteria' must be setted"));
    }

    private void ProcessSearch()
    {
      var searchingCriteria = Request.Arguments["searching criteria"] as ImapStringList;
      var tryingCharset = trialCharsets.Dequeue();

      // set charset
      var containsLiteral = false;

      searchingCriteria.Traverse(delegate(ImapString criterion) {
        if (criterion is ImapLiteralString) {
          containsLiteral = true;
          (criterion as ImapLiteralString).Charset = tryingCharset;
        }
      });

      var arguments = new List<ImapString>();

      if (containsLiteral) {
        arguments.Add("charset");
        arguments.Add(tryingCharset.WebName);
      }

      arguments.AddRange(searchingCriteria);

      if (Request.Arguments.ContainsKey("uid"))
        // UID SEARCH
        Send("uid search",
             WaitForSearch,
             arguments.ToArray());
      else
        // SEARCH
        Send("search",
             WaitForSearch,
             arguments.ToArray());
    }

    private void WaitForSearch()
    {
      ImapProtocolResponse response;

      if (!Receive(out response))
        return;

      if (response is ImapDataResponse) {
        int[] msgs;

        if (ImapDataResponseConverter.FromSearchResponse(response as ImapDataResponse, out msgs))
          messages = msgs;
      }
      else if (response is ImapCommandContinuationRequest) {
        // send divided literals
        if (!SendContinue())
          Conclude(new ImapResponse(ImapTransactionStatus.Forbidden, "sending divided literals failed"));
      }
      else if (response is ImapStatusResponse) {
        var status = response as ImapStatusResponse;

        if (status.Tag != LastTag)
          return;

        if (status.Status == ImapProtocolStatus.Ok) {
          Conclude(new ImapMessageListResponse(ImapTransactionStatus.Ok, messages, string.Format("matched {0} message(s)", messages.Length)));
        }
        else if (status.Status == ImapProtocolStatus.No) {
          // 6.4.4. SEARCH Command
          //       If the server does not support the specified [CHARSET], it MUST
          //       return a tagged NO response (not a BAD).  This response SHOULD
          //       contain the BADCHARSET response code, which MAY list the
          //       [CHARSET]s supported by the server.
          string[] supportedCharsets;

          if (ImapStatusResponseConverter.FromSearchNoResponse(status, out supportedCharsets) &&
              0 < supportedCharsets.Length) {
            // create triable Encoding instance
            foreach (var charset in supportedCharsets) {
              if (serverSupportedCharsets.Contains(charset))
                continue;

              serverSupportedCharsets.Add(charset);

              try {
                trialCharsets.Enqueue(Encoding.GetEncoding(charset));
              }
              catch (ArgumentException) {
                Trace.Info("charset {0} is not supported on this platform", charset);
              }
            }

            // try other charset
            if (0 < trialCharsets.Count)
              Transit(ProcessSearch);
            else
              Conclude(new ImapResponse(ImapTransactionStatus.BadRequest, status.Text));
          }
          else {
            ConcludeError(status);
          }
        }
        else {
          ConcludeError(status);
        }
      }
    }

    private int[] messages = new int[] {};
    private readonly List<string> serverSupportedCharsets = new List<string>();
    private readonly Queue<Encoding> trialCharsets = new Queue<Encoding>(new[] {Encoding.UTF8});
  }
}
