using System;
using System.Collections.Generic;

using Smdn.Protocols.Imap4.Protocol;

namespace Smdn.Protocols.Imap4.Transaction.Transactions {
  internal abstract class SelectTransactionBase : TransactionBase {

    // ProcessArgumentNotSetted
    //   x Conclude
    // ProcessSelect
    //   ->WaitForSelect
    //   x Conclude
    // WaitForSelect
    //   x Conclude

    protected SelectTransactionBase(string command, ImapConnection connection, ImapRequest request)
      : base(connection, request)
    {
      commandString = command.ToLower();
    }

    protected override ProcessTransactionDelegate Reset()
    {
      if (!Request.Arguments.ContainsKey("mailbox name"))
        return ProcessArgumentNotSetted;
      else
        return ProcessSelect;
    }

    private void ProcessArgumentNotSetted()
    {
      Conclude(new ImapResponse(ImapTransactionStatus.BadRequest, "arguments 'mailbox name' must be setted"));
    }

    private void ProcessSelect()
    {
      selectingMailbox = ImapMailboxManager.GetMailbox(Connection.ServerInfo, (string)Request.Arguments["mailbox name"]);

      Send(commandString,
           WaitForSelect,
           new ImapQuotedString(ImapEncoding.ToModifiedUTF7String((string)Request.Arguments["mailbox name"])));
    }

    private void WaitForSelect()
    {
      ImapProtocolResponse response;

      if (!Receive(out response))
        return;

      if (response is ImapDataResponse) {
        ImapMessageFlag flags;
        int? exists;
        int? recent;
        if (ImapDataResponseConverter.FromSelectExamineCommandResponse(response as ImapDataResponse,
                                                                       out flags,
                                                                       out exists,
                                                                       out recent)) {
          if (flags != null)          selectingMailbox.ApplicableFlags = flags;
          if (exists != null)         selectingMailbox.ExistsMessage = (int)exists;
          if (recent != null)         selectingMailbox.RecentMessage = (int)recent;
        }
      }
      else if (response is ImapStatusResponse) {
        var status = response as ImapStatusResponse;

        if (status.Status == ImapProtocolStatus.Ok) {
          bool? readOnly;
          int? unseen;
          ImapMessageFlag permanentFlags;
          int? uidNext;
          int? uidValidity;

          if (ImapStatusResponseConverter.FromSelectExamineOkResponse(status,
                                                                      out readOnly,
                                                                      out unseen,
                                                                      out permanentFlags,
                                                                      out uidNext,
                                                                      out uidValidity)) {
            if (readOnly != null)               selectingMailbox.ReadOnly = (bool)readOnly;
            if (unseen != null)                 selectingMailbox.FirstUnseen = (int)unseen;
            if (permanentFlags != null)         selectingMailbox.PermanentFlags = permanentFlags;
            if (uidNext != null)                selectingMailbox.UidNext = (int)uidNext;
            if (uidValidity != null)            selectingMailbox.UidValidity = (int)uidValidity;
          }
        }

        if (status.Tag != LastTag)
          return;

        if (status.Status == ImapProtocolStatus.Ok) {
          var description = "selected mailbox";

          if (selectingMailbox.ReadOnly) {
            description += " as read-only";
          }

          Trace.Dump(selectingMailbox);

          Conclude(new ImapMailboxResponse(ImapTransactionStatus.Ok, selectingMailbox, description));
        }
        else if (status.Status == ImapProtocolStatus.No) {
          // RFC 2193 IMAP4 Mailbox Referrals
          // http://tools.ietf.org/html/rfc2193
          // 4.1. SELECT, EXAMINE, DELETE, SUBSCRIBE, UNSUBSCRIBE, STATUS and APPEND
          //      Referrals
          ImapUri[] referrals;
          if (ImapStatusResponseConverter.FromSelectExamineNoResponse(status, out referrals))
            Conclude(new ImapMailboxReferralResponse(ImapTransactionStatus.MailboxReferral, referrals, status.Text));
          else
            ConcludeError(status);
        }
        else {
          ConcludeError(status);
        }
      }
    }

    private string commandString = string.Empty;
    private ImapMailbox selectingMailbox = null;
  }
}
