﻿using System;

using Smdn.Protocols.Imap4.Protocol;

namespace Smdn.Protocols.Imap4.Transaction.Transactions {
  internal abstract class TransactionBase : ImapTransaction {
    protected TransactionBase(ImapConnection connection, ImapRequest request)
      : base(connection, request)
    {
    }

    protected bool Aborted {
      get { return aborted; }
    }

    protected string LastTag {
      get { return lastTag; }
    }

    public override void Abort()
    {
      aborted = true;

      base.Abort();
    }

    protected bool Receive(out ImapProtocolResponse response)
    {
      response = null;

      if (aborted)
        return false;

      response = Connection.TryReceiveResponse();

      if (response is ImapStatusResponse) {
        var status = response as ImapStatusResponse;
        string message;
        if (ImapStatusResponseConverter.FromParse(status, out message)) {
          // PARSE
          Trace.Info("[PARSE] {0}", message);
        }
        else if (ImapStatusResponseConverter.FromAlert(status, out message)) {
          // ALERT
          Trace.Info("[ALERT] {0}", message);
          Console.Error.WriteLine("[ALERT] {0}", message);
        }
      }

      return (response != null);
    }

    protected string Send(string command, ImapTransaction.ProcessTransactionDelegate nextProcess, params ImapString[] arguments)
    {
      if (aborted)
        return string.Empty;

      var comm = Connection.CreateCommand(command, arguments);

      Send(comm);

      Transit(nextProcess);

      lastTag = comm.Tag;

      return lastTag;
    }

    protected void SendContinue(params ImapString[] arguments)
    {
      if (aborted)
        return;

      var comm = Connection.CreateContinuingCommand(arguments);

      Send(comm);
    }

    protected bool SendContinue()
    {
      if (aborted)
        return false;

      if (dividedCommandContext == null)
        return false;

      dividedCommandContext = Connection.SendCommand(dividedCommandContext);

      return true;
    }

    private void Send(ImapProtocolCommand command)
    {
      if (aborted)
        return;

      dividedCommandContext = Connection.SendCommand(command);
    }

    protected void ProcessIncalable()
    {
      Conclude(new ImapResponse(ImapTransactionStatus.Incapable, "required command is incapable"));
    }

    protected void Conclude(ImapResponse response)
    {
      if (aborted)
        return;

      this.Response = response;

      Transit(null);
    }

    protected void ConcludeError(ImapStatusResponse response)
    {
      if (aborted)
        return;

      if (response.Status ==ImapProtocolStatus.No) {
        Conclude(new ImapResponse(ImapTransactionStatus.Forbidden, response.Text));
      }
      else if (response.Status == ImapProtocolStatus.Bad) {
        Conclude(new ImapResponse(ImapTransactionStatus.InternalServerError, response.Text));
      }
      else if (response.Status == ImapProtocolStatus.Bye) {
        Conclude(new ImapResponse(ImapTransactionStatus.Disconnecting, response.Text));
      }
      else {
        Conclude(new ImapResponse(ImapTransactionStatus.BadRequest, response.Text));
      }
    }

    private bool aborted = false;
    private string lastTag = null;
    private ImapConnection.DividedCommandContext dividedCommandContext = null;
  }
}
