using System;
using System.Diagnostics;
using System.IO;
using System.Text;
using System.Reflection;
using System.Threading;

using Smdn.Protocols.Imap4.Protocol;
using Smdn.Protocols.Imap4.Transaction;

namespace Smdn.Protocols.Imap4 {
  internal static class Trace {
    private static readonly TraceSource     traceSource = new TraceSource(Assembly.GetExecutingAssembly().GetName().Name, SourceLevels.Verbose);
    private static readonly TraceSource dumpTraceSource = new TraceSource(Assembly.GetExecutingAssembly().GetName().Name + "#Dump", SourceLevels.Verbose);

    private static string Now()
    {
      return string.Format("{0:s}.{1:d4} ", DateTime.Now, DateTime.Now.Millisecond);
    }

    [Conditional("TRACE")]
    public static void Info(string format, params object[] args)
    {
      traceSource.TraceEvent(TraceEventType.Information, Thread.CurrentThread.ManagedThreadId, Now() + format, args);
    }

    [Conditional("TRACE")]
    public static void Verbose(string format, params object[] args)
    {
      traceSource.TraceEvent(TraceEventType.Verbose, Thread.CurrentThread.ManagedThreadId, Now() + format, args);
    }

    [Conditional("TRACE")]
    public static void Error(string format, params object[] args)
    {
      traceSource.TraceEvent(TraceEventType.Error, Thread.CurrentThread.ManagedThreadId, Now() + format, args);
    }

    [Conditional("TRACE")]
    public static void Log(Exception ex)
    {
      traceSource.TraceEvent(TraceEventType.Error, Thread.CurrentThread.ManagedThreadId, "Σ（＞ω＜）{0}{1}", Environment.NewLine, ex);
    }

    [Conditional("TRACE")]
    public static void Log(ImapConnection connection, ImapProtocolResponse response)
    {
      var prefix = string.Format("CID:{0} 　Σ（∵）[IMAP4rev1]", connection.Id);

      var reader = new StringReader(response.Response);

      for (;;) {
        var line = reader.ReadLine();

        if (line == null)
          break;

        Verbose("{0} {1}", prefix, line);
      }
    }

    [Conditional("TRACE")]
    public static void Log(ImapConnection connection, byte[] command)
    {
      Log(connection, command, 0, command.Length);
    }

    [Conditional("TRACE")]
    public static void Log(ImapConnection connection, byte[] command, int start, int count)
    {
      var prefix = string.Format("CID:{0} （＞ω＜）[IMAP4rev1]", connection.Id);

      var reader = new StringReader(ImapEncoding.Encoding8Bit.GetString(command, start, count));

      for (;;) {
        var line = reader.ReadLine();

        if (line == null)
          break;

        Verbose("{0} {1}", prefix, line);
      }
    }

    [Conditional("TRACE")]
    public static void Log(ImapRequest request, ImapTransaction transaction)
    {
      var param = string.Empty;

      foreach (var p in request.Arguments) {
        var val = p.Value;

        if (p.Key == "password")
          val = "****";

        param += string.Format(" '{0}'=>{{{1}}}", p.Key, val);
      }

      Verbose("CID:{0} SeqID:{1} [ImapTransaction] {2}{3}",
              transaction.Connection.Id,
              transaction.SequenceId,
              request.Method,
              param);
    }

    [Conditional("TRACE")]
    public static void Log(ImapResponse response, ImapTransaction transaction)
    {
      Verbose("CID:{0} SeqID:{1} [ImapTransaction] {2} {3} '{4}'",
              transaction.Connection.Id,
              transaction.SequenceId,
              transaction.Request.Method,
              response.Status,
              response.Description);
    }

    [Conditional("DUMP")]
    public static void DumpInfo(string format, params object[] args)
    {
      dumpTraceSource.TraceEvent(TraceEventType.Information, Thread.CurrentThread.ManagedThreadId, Now() + format, args);
    }

    [Conditional("DUMP")]
    public static void DumpVerbose(string format, params object[] args)
    {
      dumpTraceSource.TraceEvent(TraceEventType.Verbose, Thread.CurrentThread.ManagedThreadId, Now() + format, args);
    }

    [Conditional("DUMP")]
    public static void Dump(byte[] buffer)
    {
      Dump(buffer, 0, buffer.Length);
    }

    [Conditional("DUMP")]
    public static void Dump(byte[] buffer, int start, int count)
    {
      var sb = new StringBuilder();
      var end = start + count;

      for (var index = start; index < end; index++) {
        if (index % 0x20 == 0)
          sb.AppendLine();
        sb.AppendFormat("{0:x2} ", buffer[index]);
      }

      DumpVerbose("Length:{0}{1}{2}{3}",
              count,
              Environment.NewLine,
              ImapEncoding.Encoding8Bit.GetString(buffer, start, count),
              sb.ToString());
    }

    [Conditional("DUMP")]
    public static void Dump(ImapProtocolResponse response)
    {
      if (response is ImapDataResponse) {
        DumpInfo("[ImapResponseData] {0}", response);

        foreach (var d in (response as ImapDataResponse).Data) {
          Dump(d, 0);
        }
      }
      else if (response is ImapStatusResponse) {
        DumpInfo("[ImapStatusResponse] {0}", response);
      }
      else if (response is ImapCommandContinuationRequest) {
        DumpInfo("[ImapCommandContinuationRequest] {0}", response);
      }
    }

    [Conditional("DUMP")]
    private static void Dump(ImapData data, int indent)
    {
      DumpVerbose("[ImapData] {0}{1}", new string(' ', indent * 4), data);

      if (data.Format == ImapDataFormat.List) {
        foreach (var l in data.List) {
          Dump(l, indent + 1);
        }
      }
    }

    [Conditional("DUMP")]
    public static void Dump(ImapMailbox mailbox)
    {
      var prefix = string.Empty; // new string(' ', indent * 4);

      if (mailbox == null) {
        DumpVerbose("{0}[ImapMailbox] null", prefix);
        return;
      }

      DumpVerbose("{0}[ImapMailbox] begin", prefix);
      DumpVerbose("{0}  Name={1}", prefix, mailbox.Name);
      DumpVerbose("{0}  HierarchyDelimiter={1}", prefix, mailbox.HierarchyDelimiter);
      DumpVerbose("{0}  ReadOnly={1}", prefix, mailbox.ReadOnly);
      DumpVerbose("{0}  Flags={1}", prefix, mailbox.Flags);
      DumpVerbose("{0}  ApplicableFlags={1}", prefix, mailbox.ApplicableFlags);
      DumpVerbose("{0}  PermanentFlags={1}", prefix, mailbox.PermanentFlags);
      DumpVerbose("{0}  ExistsMessage={1}", prefix, mailbox.ExistsMessage);
      DumpVerbose("{0}  RecentMessage={1}", prefix, mailbox.RecentMessage);
      DumpVerbose("{0}  UnseenMessage={1}", prefix, mailbox.UnseenMessage);
      DumpVerbose("{0}  FirstUnseen={1}", prefix, mailbox.FirstUnseen);
      DumpVerbose("{0}  UidNext={1}", prefix, mailbox.UidNext);
      DumpVerbose("{0}  UidValidity={1}", prefix, mailbox.UidValidity);
      DumpVerbose("{0}[ImapMailbox] end", prefix);
    }

    [Conditional("DUMP")]
    public static void Dump(ImapMessage message, int indent)
    {
      var prefix = new string(' ', indent * 4);

      if (message == null) {
        DumpVerbose("{0}[ImapMessage] null", prefix);
        return;
      }

      DumpVerbose("{0}[ImapMessage] begin", prefix);
      DumpVerbose("{0}  Mailbox={1}", prefix, message.Mailbox);
      DumpVerbose("{0}  Sequence={1}", prefix, message.Sequence);
      DumpVerbose("{0}  Uid={1}", prefix, message.Uid);
      DumpVerbose("{0}  BodySize={1}", prefix, message.BodySize);
      DumpVerbose("{0}  InternalDate={1}", prefix, message.InternalDate);
      DumpVerbose("{0}  Flags={1}", prefix, message.Flags);
      DumpVerbose("{0}  Envelope:", prefix); Dump(message.Envelope, indent + 1);
      DumpVerbose("{0}  BodyStructure:", prefix); Dump(message.BodyStructure, indent + 1);
      DumpVerbose("{0}  Body:{1}", prefix, message.Body);
      DumpVerbose("{0}[ImapMessage] end", prefix);
    }

    [Conditional("DUMP")]
    public static void Dump(ImapBodyStructure bodyStructure, int indent)
    {
      var prefix = new string(' ', indent * 4);

      if (bodyStructure == null) {
        DumpVerbose("{0}[ImapBodyStructure] null", prefix);
        return;
      }

      DumpVerbose("{0}[ImapBodyStructure] begin", prefix);
      DumpVerbose("{0}  Type={1}/{2}", prefix, bodyStructure.Type, bodyStructure.Subtype);
      DumpVerbose("{0}  Envelope:", prefix); Dump(bodyStructure.Envelope, indent + 1);
      foreach (var pair in bodyStructure.Parameters) {
        DumpVerbose("{0}  Parameters {1}={2}", prefix, pair.Key, pair.Value);
      }
      DumpVerbose("{0}  Encoding={1}", prefix, bodyStructure.Encoding);
      DumpVerbose("{0}  Id={1}", prefix, bodyStructure.Id);
      DumpVerbose("{0}  Description={1}", prefix, bodyStructure.Description);
      DumpVerbose("{0}  Size={1}", prefix, bodyStructure.Size);
      DumpVerbose("{0}  Lines={1}", prefix, bodyStructure.Lines);
      DumpVerbose("{0}  Nested:", prefix);
      foreach (var nested in bodyStructure.Nested) {
        Dump(nested, indent + 1);
      }

      DumpVerbose("{0}[ImapBodyStructure] end", prefix);
    }

    [Conditional("DUMP")]
    public static void Dump(ImapEnvelope envelope, int indent)
    {
      var prefix = new string(' ', indent * 4);

      if (envelope == null) {
        DumpVerbose("{0}[ImapEnvelope] null", prefix);
        return;
      }

      DumpVerbose("{0}[ImapEnvelope] begin", prefix);
      DumpVerbose("{0}  Date={1}", prefix, envelope.Date);
      DumpVerbose("{0}  Subject={1}", prefix, envelope.Subject);
      DumpVerbose("{0}  From:", prefix); Dump(envelope.From, indent + 1);
      DumpVerbose("{0}  Sender:", prefix); Dump(envelope.Sender, indent + 1);
      DumpVerbose("{0}  ReplyTo:", prefix); Dump(envelope.ReplyTo, indent + 1);
      DumpVerbose("{0}  To:", prefix); Dump(envelope.To, indent + 1);
      DumpVerbose("{0}  Cc:", prefix); Dump(envelope.Cc, indent + 1);
      DumpVerbose("{0}  Bcc:", prefix); Dump(envelope.Bcc, indent + 1);
      DumpVerbose("{0}  InReplyTo={1}", prefix, envelope.InReplyTo);
      DumpVerbose("{0}  MessageId={1}", prefix, envelope.MessageId);
      DumpVerbose("{0}[ImapEnvelope] end", prefix);
    }

    [Conditional("DUMP")]
    public static void Dump(ImapAddress address, int indent)
    {
      var prefix = new string(' ', indent * 4);

      if (address == null) {
        DumpVerbose("{0}[ImapAddress] null", prefix);
        return;
      }

      DumpVerbose("{0}[ImapAddress] begin", prefix);
      DumpVerbose("{0}  Name={1}", prefix, address.Name);
      DumpVerbose("{0}  AtDomailList={1}", prefix, address.AtDomailList);
      DumpVerbose("{0}  Mailbox={1}", prefix, address.Mailbox);
      DumpVerbose("{0}  Host={1}", prefix, address.Host);
      DumpVerbose("{0}[ImapAddress] end", prefix);
    }
  }
}
