using System;
using System.Collections.Generic;
using System.IO;
using System.Net.Security;
using System.Security.Authentication;
using System.Security.Cryptography.X509Certificates;

using Smdn.Protocols.Pop3.Session;
using Smdn.Protocols.Pop3.Transaction;

namespace Smdn.Protocols.Pop3.Client {
  // Layers:
  //   Client.PopClient
  // * Client.PopClientAgent
  //   Session.PopSessionAgent
  //   Session.PopSession
  //   Transaction.PopTransactionAgent
  //   Transaction.PopTransaction/Transaction.PopRequest/Transaction.PopResponse
  //   Protocol.PopConnection/Protocol.PopProtocolCommand/Protocol.PopProtocolResponse
  //   (POP3)

  public class PopClientAgent : IDisposable {
    public X509Certificate2Collection CertCollection {
      get { return agent.CertCollection; }
    }
    
    public RemoteCertificateValidationCallback ServerCertificateValidationCallback {
      get { return agent.ServerCertificateValidationCallback; }
      set { agent.ServerCertificateValidationCallback = value; }
    }

    public PopClientAgent()
    {
    }

    public void Initialize()
    {
      if (agent != null)
        throw new InvalidOperationException("already initialized");

      Trace.Info("PopClientAgent initializing");

      agent = new PopSessionAgent();
      agent.CreateAuthenticatedStreamCallback = CreateAuthenticatedStream;
      agent.UpgradeAuthenticatedStreamCallback = UpgradeAuthenticatedStream;
      agent.Initialize();

      Trace.Info("PopClientAgent initialized");
    }

    ~PopClientAgent()
    {
      Dispose();
      GC.SuppressFinalize(this);
    }

    public void Dispose()
    {
      if (disposed)
        return;

      Trace.Info("PopClientAgent disposing");

      if (agent != null) {
        PopClient[] existClients;

        lock (clients) {
          existClients = clients.ToArray();
        }

        foreach (var client in existClients) {
          if (client.Session != null)
            Logout(client);
        }

        lock (clients) {
          clients.Clear();
        }

        agent.Dispose();
        agent = null;
      }

      Trace.Info("PopClientAgent disposed");

      disposed = true;
    }

    public PopClient CreateClient()
    {
      var client = new PopClient(this);

      lock (clients) {
        clients.Add(client);
      }

      return client;
    }

    /*
     * methods for authentication and session management 
     */

    public void Connect(PopClient client)
    {
      if (client.Session != null)
        throw new InvalidOperationException("already connected");

      if (client.HostUri == null)
        throw new InvalidOperationException("HostUri must be set");

      var session = agent.GetExistSession(client.HostUri);

      if (session != null)
        throw new InvalidOperationException("other client has already started requested session");

      if (!agent.Connect(client.HostUri, out session))
        throw new PopClientExcetion("connect failed");

      client.Session = session;
    }

    /// <summary>sends USER/PASS, APOP, or AUTH, and CAPA request</summary>
    public void Authenticate(PopClient client)
    {
      try {
        var session = client.Session;

        if (session == null)
          throw new InvalidOperationException("not connected");

        for (var tryTls = 0; tryTls < 2; tryTls++) {
          // try CAPA
          PopCapability capability; // no use

          if (agent.Capa(session, out capability)) {
            if (client.TryTls && tryTls == 0 &&
                session.ServerInfo.Capabilities.Has(PopKnownCapabilities.Stls))
              if (agent.Stls(session))
                // conprehend capability again
                continue;
          }

          break;
        }

        if (session.HostUri.Authentication == PopAuthenticationType.Apop) {
          // try APOP first
          if (agent.Apop(session, client.Password))
            return;

          // try AUTH
          if (agent.Auth(session, client.Password))
            return;
        }
        else {
          // try AUTH first
          if (agent.Auth(session, client.Password))
            return;

          // try APOP
          if (agent.Apop(session, client.Password))
            return;
        }

        // try USER/PASS
        if (!agent.UserPass(session, client.Password)) {
          agent.Quit(session);
          client.Session = null;
          throw new PopClientExcetion("authentication was refused");
        }
      }
      finally {
        CheckConnectionAlive(client);

        if (client.Session != null)
          client.Mailbox = new PopMailbox();
      }
    }

    /// <summary>sends QUIT request</summary>
    public void Logout(PopClient client)
    {
      RejectNonLoggedIn(client);

      if (!agent.Quit(client.Session))
        throw new PopClientExcetion("logout failed");

      client.Session = null;
      client.Mailbox = null;
    }

    /*
     * methods for requesting 
     */

    /// <summary>sends LIST request</summary>
    public PopMessage[] ListMessages(PopClient client)
    {
      RejectNonLoggedIn(client);

      try {
        PopScanList[] scanLists;

        if (!agent.List(client.Session, out scanLists))
          throw new PopClientExcetion("listing messages failed");

        // update mailbox and message status
        var mailbox = client.Mailbox;

        mailbox.ManagedMessages.Clear();

        foreach (var scanList in scanLists) {
          var message = new PopMessage(scanList);
          message.Mailbox = mailbox;
          mailbox.ManagedMessages.Add(message.Number, message);
        }

        mailbox.ExistsMessage = scanLists.Length;

        return mailbox.Messages;
      }
      finally {
        CheckConnectionAlive(client);
      }
    }

    /// <summary>sends UIDL request</summary>
    public string GetUid(PopClient client, PopMessage message)
    {
      RejectNonLoggedIn(client);

      try {
        PopUniqueIdList uidList;

        if (!agent.Uidl(client.Session, message, out uidList))
          throw new PopClientExcetion("UID listing failed");

        client.Mailbox.ManagedMessages[uidList.MessageNumber].Uid = uidList.Uid;

        return uidList.Uid;
      }
      finally {
        CheckConnectionAlive(client);
      }
    }

    /// <summary>sends STAT request</summary>
    public void UpdateMailboxStatus(PopClient client)
    {
      RejectNonLoggedIn(client);

      try {
        PopDropList dropList;

        if (!agent.Stat(client.Session, out dropList))
          throw new PopClientExcetion("updating mailbox status failed");

        // update mailbox status
        client.Mailbox.ExistsMessage = dropList.Messages;
        client.Mailbox.TotalOctets = dropList.Octets;
      }
      finally {
        CheckConnectionAlive(client);
      }
    }

    /// <summary>sends RETR request</summary>
    public PopMessage Retrieve(PopClient client, PopMessage message)
    {
      RejectNonLoggedIn(client);

      try {
        PopMessage retrieved;

        if (!agent.Retr(client.Session, message.Number, out retrieved))
          throw new PopClientExcetion("retrieving a message body failed");

        // update message status
        var managed = client.Mailbox.ManagedMessages[message.Number];

        managed.Body = retrieved.Body;

        return managed;
      }
      finally {
        CheckConnectionAlive(client);
      }
    }

    /// <summary>sends RETR and DELE request</summary>
    public PopMessage RetrieveAndDelete(PopClient client, PopMessage message)
    {
      var retrieved = Retrieve(client, message);

      Delete(client, message);

      return retrieved;
    }

    /// <summary>sends TOP request</summary>
    public PopMessage RetrieveHeader(PopClient client, PopMessage message)
    {
      RejectNonLoggedIn(client);

      try {
        PopMessage retrieved;

        if (!agent.Top(client.Session, message.Number, 0, out retrieved))
          if (!agent.Retr(client.Session, message.Number, out retrieved)) // try RETR
            throw new PopClientExcetion("retrieving a message header failed");

        // update message status
        var managed = client.Mailbox.ManagedMessages[message.Number];

        managed.Body = retrieved.Body;

        return managed;
      }
      finally {
        CheckConnectionAlive(client);
      }
    }

    /// <summary>sends DELE request</summary>
    public void Delete(PopClient client, PopMessage message)
    {
      RejectNonLoggedIn(client);

      try {
        if (!agent.Dele(client.Session, message))
          throw new PopClientExcetion("deletion failed");

        // update mailbox and message status
        client.Mailbox.ManagedMessages.Remove(message.Number);
        client.Mailbox.Deleted = true;

        UpdateMailboxStatus(client);
      }
      finally {
        CheckConnectionAlive(client);
      }
    }

    /// <summary>sends RSET request</summary>
    public void CancelDeletion(PopClient client)
    {
      RejectNonLoggedIn(client);

      if (!client.Mailbox.Deleted)
        return;

      try {
        if (!agent.Rset(client.Session))
          throw new PopClientExcetion("canceling deletion failed");

        // update mailbox and message status
        client.Mailbox.Deleted = false;

        UpdateMailboxStatus(client);

        ListMessages(client);
      }
      finally {
        CheckConnectionAlive(client);
      }
    }

    /// <summary>sends NOOP request</summary>
    public void NoOp(PopClient client)
    {
      RejectNonLoggedIn(client);

      try {
        // continue whether error occurred
        agent.NoOp(client.Session);
      }
      finally {
        CheckConnectionAlive(client);
      }
    }

    /*
     * private methods 
     */

    private void RejectNonLoggedIn(PopClient client)
    {
      if (client.Session == null)
        throw new InvalidOperationException("not logged in");
    }

    private void CheckConnectionAlive(PopClient client)
    {
      if (client.Session == null)
        return;

      if (client.Session.State == PopSessionState.NotConnected) {
        client.Session = null;

        throw new PopDisconnectedExcetion("session was disconnected by server");
      }
    }

    private PopClient GetExistClientFromUri(PopUri uri)
    {
      PopClient[] existClients;

      lock (clients) {
        existClients = clients.ToArray();
      }

      foreach (var client in existClients) {
        if (client.HostUri.UserAtHost == uri.UserAtHost)
          return client;
      }

      return null;
    }

    private Stream CreateAuthenticatedStream(PopUri uri, Stream baseStream)
    {
      var client = GetExistClientFromUri(uri);

      if (client != null && client.UseSsl)
        return UpgradeAuthenticatedStream(uri, baseStream);
      else
        return baseStream;
    }

    private Stream UpgradeAuthenticatedStream(PopUri uri, Stream baseStream)
    {
      SslStream sslStream = null;

      if (agent.ServerCertificateValidationCallback == null)
        sslStream = new SslStream(baseStream, false);
      else
        sslStream = new SslStream(baseStream, false, agent.ServerCertificateValidationCallback);

      sslStream.AuthenticateAsClient(uri.Host, agent.CertCollection, SslProtocols.Default, true);

      Trace.Info("created SSL stream for {0}", uri);

      return sslStream;
    }

    private bool disposed = false;
    private PopSessionAgent agent = null;
    private readonly List<PopClient> clients = new List<PopClient>();
  }
}
